from __future__ import annotations

import os
import subprocess
from typing import Optional

import typer

from gamspy.exceptions import ValidationError
import gamspy.utils as utils

from . import install, list, retrieve, run, show, uninstall

app = typer.Typer(
    rich_markup_mode="rich",
    context_settings={"help_option_names": ["-h", "--help"]},
)
app.add_typer(install.app, name="install")
app.add_typer(list.app, name="list")
app.add_typer(retrieve.app, name="retrieve")
app.add_typer(run.app, name="run")
app.add_typer(show.app, name="show")
app.add_typer(uninstall.app, name="uninstall")


def version_callback(value: bool):
    if value:
        import gams

        import gamspy

        print(f"GAMSPy version: {gamspy.__version__}")
        print(f"GAMS version: {gams.__version__}")

        try:
            import gamspy_base

            print(f"gamspy_base version: {gamspy_base.__version__}")
        except ModuleNotFoundError:
            ...

        raise typer.Exit()


@app.callback()
def callback(
    version: Optional[bool] = typer.Option(
        None,
        "-v",
        "--version",
        help="Shows the version of gamspy, gamsapi, and gamspy_base.",
        callback=version_callback,
    ),
) -> None:
    """
    GAMSPy CLI - The [bold]gamspy[/bold] command line app. 😎

    Install solvers and licenses, run MIRO apps, and more.

    Read more in the docs: [link=https://gamspy.readthedocs.io/en/latest/cli/index.html]https://gamspy.readthedocs.io/en/latest/cli/index.html[/link].
    """
    ...

@app.command(short_help="To probe node information.")
def probe(
    json_out: Optional[str] = typer.Option(
        None,
        "--json-out", "-j",
        help="Output path for the JSON file."
    ),
):
    gamspy_base_dir = utils._get_gamspy_base_directory()
    process = subprocess.run(
        [os.path.join(gamspy_base_dir, "gamsprobe")],
        text=True,
        capture_output=True,
    )

    if process.returncode:
        raise ValidationError(process.stderr)

    print(process.stdout)

    if json_out:
        with open(json_out, "w") as file:
            file.write(process.stdout)

def main():
    """
    Entry point for gamspy command line application.
    """
    app()



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

import gamspy as gp
import gamspy.formulations.nn.utils as utils
from gamspy.exceptions import ValidationError
from gamspy.math import dim

if TYPE_CHECKING:
    from gamspy import Parameter, Variable


class Linear:
    """
    Formulation generator for Linear layer in GAMS.

    Parameters
    ----------
    container : Container
        Container that will contain the new variable and equations.
    in_features : int
        Input feature size
    out_features : int
        Output feature size
    bias : bool = True
        Should bias be added after linear transformation, by Default: True
    name_prefix : str | None
        Prefix for generated GAMSPy symbols, by default None which means
        random prefix. Using the same name_prefix in different formulations causes name
        conflicts. Do not use the same name_prefix again.

    Examples
    --------
    >>> import gamspy as gp
    >>> import numpy as np
    >>> from gamspy.math import dim
    >>> m = gp.Container()
    >>> l1 = gp.formulations.Linear(m, 128, 64)
    >>> w = np.random.rand(64, 128)
    >>> b = np.random.rand(64)
    >>> l1.load_weights(w, b)
    >>> x = gp.Variable(m, "x", domain=dim([10, 128]))
    >>> y, set_y = l1(x)
    >>> [d.name for d in y.domain]
    ['DenseDim10_1', 'DenseDim64_1']

    """

    def __init__(
        self,
        container: gp.Container,
        in_features: int,
        out_features: int,
        bias: bool = True,
        name_prefix: str | None = None,
    ):
        if not isinstance(in_features, int) or in_features <= 0:
            raise ValidationError("in_features must be a positive integer")

        if not isinstance(out_features, int) or out_features <= 0:
            raise ValidationError("out_features must be a positive integer")

        if not isinstance(bias, bool):
            raise ValidationError("bias must be a boolean")

        self.container = container
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias
        self._state = 0
        self.weight: Parameter | Variable | None = None
        self.weight_array = None
        self.bias: Parameter | Variable | None = None
        self.bias_array = None

        if name_prefix is None:
            name_prefix = gp.utils._get_unique_name()

        self._name_prefix = name_prefix

    def load_weights(
        self, weight: np.ndarray, bias: np.ndarray | None = None
    ) -> None:
        """
        Mark Linear as parameter and load weights from NumPy arrays.
        After this is called `make_variable` cannot be called. Use this
        when you already have the weights of your Linear layer.

        Parameters
        ----------
        weight : np.ndarray
               Linear layer weights in shape (out_features x in_features)
        bias : np.ndarray | None
               Linear layer bias in shape (out_features, ), only required when
               bias=True during initialization

        """
        if self._state == 2:
            raise ValidationError(
                "load_weights cannot be used after calling make_variable"
            )

        if self.use_bias is False and bias is not None:
            raise ValidationError(
                "bias must be None since bias was set to False during initialization"
            )

        if self.use_bias is True and bias is None:
            raise ValidationError("bias must be provided")

        if len(weight.shape) != 2:
            raise ValidationError(
                f"expected 2D input for weight (got {len(weight.shape)}D input)"
            )

        expected_shape = (
            self.out_features,
            self.in_features,
        )
        if weight.shape != expected_shape:
            raise ValidationError(
                f"weight expected to be in shape {expected_shape}"
            )

        if bias is not None:
            if len(bias.shape) != 1:
                raise ValidationError(
                    f"expected 1D input for bias (got {len(bias.shape)}D input)"
                )

            if bias.shape[0] != self.out_features:
                raise ValidationError(
                    f"bias expected to be in shape ({self.out_features},)"
                )

        if self.weight is None:
            self.weight = gp.Parameter(
                self.container,
                name=utils._generate_name("p", self._name_prefix, "weight"),
                domain=dim(expected_shape),
                records=weight,
            )
        else:
            self.weight.setRecords(weight)
        self.weight_array = weight

        if self.use_bias:
            if self.bias is None:
                self.bias = gp.Parameter(
                    self.container,
                    name=utils._generate_name("p", self._name_prefix, "bias"),
                    domain=dim([self.out_features]),
                    records=bias,
                )
            else:
                self.bias.setRecords(bias)

            self.bias_array = bias

        self._state = 1

    def make_variable(self) -> None:
        """
        Mark Linear layer as variable. After this is called `load_weights`
        cannot be called. Use this when you need to learn the weights
        of your linear layer in your optimization model.

        This does not initialize the weights, it is highly recommended
        that you set initial values to `weight` and `bias` variables.
        """
        if self._state == 1:
            raise ValidationError(
                "make_variable cannot be used after calling load_weights"
            )

        expected_shape = (
            self.out_features,
            self.in_features,
        )

        if self.weight is None:
            self.weight = gp.Variable(
                self.container,
                name=utils._generate_name("v", self._name_prefix, "weight"),
                domain=dim(expected_shape),
            )

        if self.use_bias and self.bias is None:
            self.bias = gp.Variable(
                self.container,
                name=utils._generate_name("v", self._name_prefix, "bias"),
                domain=dim([self.out_features]),
            )

        self._state = 2

    def __call__(
        self, input: gp.Parameter | gp.Variable, propagate_bounds: bool = True
    ) -> tuple[gp.Variable, list[gp.Equation]]:
        """
        Forward pass your input, generate output and equations required for
        calculating the linear transformation. If `propagate_bounds` is True,
        the `input` is of type variable, and `load_weights` was called, then
        the bounds of the input are propagated to the output.

        Parameters
        ----------
        input : gp.Parameter | gp.Variable
                input to the linear layer, must be in shape
                (* x in_features)
        propagate_bounds : bool = True
                If True, propagate bounds of the input to the output.
                Otherwise, the output variable is unbounded.
        """
        if not isinstance(propagate_bounds, bool):
            raise ValidationError("propagate_bounds should be a boolean.")

        if self.weight is None:
            raise ValidationError(
                "You must call load_weights or make_variable first before using the Linear"
            )

        if len(input.domain) == 0:
            raise ValidationError(
                "expected an input with at least 1 dimension"
            )

        if len(input.domain[-1]) != self.in_features:
            raise ValidationError("in_features does not match")

        expr = input @ self.weight.t()

        if self.bias is not None:
            expr = expr + self.bias[expr.domain[-1]]

        out = gp.Variable(
            self.container,
            name=utils._generate_name("v", self._name_prefix, "output"),
            domain=expr.domain,
        )

        set_out = gp.Equation(
            self.container,
            name=utils._generate_name("e", self._name_prefix, "set_output"),
            domain=out.domain,
        )

        set_out[...] = out == expr

        # If propagate_bounds is True, weight is a parameter and input is a variable,
        # we will propagate the bounds of the input to the output
        if (
            propagate_bounds
            and self._state == 1
            and isinstance(input, gp.Variable)
        ):
            x_bounds = gp.Parameter(
                self.container,
                name=utils._generate_name(
                    "p", self._name_prefix, "input_bounds"
                ),
                domain=dim([2, *input.shape]),
            )
            x_bounds[("0",) + tuple(input.domain)] = input.lo[...]
            x_bounds[("1",) + tuple(input.domain)] = input.up[...]

            # If the bounds are all zeros (None in GAMSPy parameters);
            # we skip matrix multiplication as it will result in zero values
            if x_bounds.records is None:
                out_bounds_array = np.zeros(out.shape)

                if self.use_bias:
                    out_bounds_array = out_bounds_array + self.bias_array

                out_bounds = gp.Parameter(
                    self.container,
                    name=utils._generate_name(
                        "p", self._name_prefix, "output_bounds"
                    ),
                    domain=dim(out.shape),
                    records=out_bounds_array,
                )
                out.lo[...] = out_bounds
                out.up[...] = out_bounds

                return out, [set_out]

            x_lb, x_ub = x_bounds.toDense()

            # To deal with infinity values in the input bounds, we convert them into complex numbers
            # where if the value is -inf, we convert it to 0 - 1j
            # and if the value is inf, we convert it to 0 + 1j
            x_lb = np.where(x_lb == -np.inf, 0 - 1j, x_lb)
            x_ub = np.where(x_ub == np.inf, 0 + 1j, x_ub)

            # get the positive and negative weights separately
            w_pos = np.maximum(self.weight_array, 0)
            w_neg = np.minimum(self.weight_array, 0)

            lo_out = (x_lb @ w_pos.T) + (x_ub @ w_neg.T)
            up_out = (x_ub @ w_pos.T) + (x_lb @ w_neg.T)

            def _decode_complex_number(z: np.complex128) -> float:
                """
                Decode complex number to real number.
                5 + 0j -> 5
                3 + 1j -> inf
                7 - 3j -> -inf
                """
                # If imaginary part is zero, return real part
                if z.imag == 0:
                    return z.real
                # If imaginary part is positive, return positive infinity
                elif z.imag > 0:
                    return np.inf
                # If imaginary part is negative, return negative infinity
                else:
                    return -np.inf

            lo_out = np.vectorize(_decode_complex_number)(lo_out)
            up_out = np.vectorize(_decode_complex_number)(up_out)

            if self.use_bias:
                lo_out = lo_out + self.bias_array
                up_out = up_out + self.bias_array

            out_bounds_array = np.stack([lo_out, up_out], axis=0)

            out_bounds = gp.Parameter(
                self.container,
                name=utils._generate_name(
                    "p", self._name_prefix, "output_bounds"
                ),
                domain=dim([2, *out.shape]),
                records=out_bounds_array,
            )

            out.lo[...] = out_bounds[("0",) + tuple(out.domain)]
            out.up[...] = out_bounds[("1",) + tuple(out.domain)]

        return out, [set_out]



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._algebra.expression as expression
import gamspy.utils as utils
from gamspy._symbols.implicits.implicit_symbol import ImplicitSymbol
from gamspy._symbols.symbol import Symbol
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Set
    from gamspy._algebra.expression import Expression
    from gamspy._symbols.symbol import Symbol
    from gamspy._types import OperableType


class MathOp:
    def __init__(
        self,
        op_name: str,
        elements: tuple,
        safe_cancel: bool = False,
    ):
        self.op_name = op_name
        self.elements = list(elements)
        self.safe_cancel = safe_cancel

    def gamsRepr(self) -> str:
        operands_str = ",".join([_stringify(elem) for elem in self.elements])
        return f"{self.op_name}({operands_str})"

    def latexRepr(self) -> str:
        """
        Representation of this MathOp in Latex.

        Returns
        -------
        str
        """
        op_map = {
            "sqrt": "\\sqrt",
            "floor": "\\floor",
            "ceil": "\\lceil",
            "abs": "\\lvert",
        }

        operands_str = ",".join([_stringify(elem) for elem in self.elements])
        if self.op_name in op_map:
            return f"{op_map[self.op_name]}{{{operands_str}}}"

        return f"{self.op_name}({operands_str})"

    def __str__(self):
        return self.gamsRepr()

    def __len__(self):
        return len(self.gamsRepr())


def _stringify(x: str | int | float | Symbol | ImplicitSymbol):
    if isinstance(x, (int, float)):
        x = utils._map_special_values(x)

        return str(x)
    elif isinstance(x, str):
        return f'"{x}"'

    return x.gamsRepr()  # type: ignore


def abs(x: OperableType) -> Expression:
    """
    Absolute value of ``x`` (i.e. ``|x|``)

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import abs
    >>> m = Container()
    >>> a = Parameter(m, "a", records=-3.8)
    >>> b = Parameter(m, "b")
    >>> b[...] = abs(a)
    >>> b.toValue()
    np.float64(3.8)

    """
    return expression.Expression(None, MathOp("abs", (x,)), None)


def ceil(x: OperableType) -> Expression:
    """
    The smallest integer greater than or equal to ``x`` (i.e. ``ceil(4.1)`` returns ``5``)

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import ceil
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.2)
    >>> b = Parameter(m, "b")
    >>> b[...] = ceil(a)
    >>> b.toValue()
    np.float64(4.0)

    """
    return expression.Expression(None, MathOp("ceil", (x,)), None)


def div(dividend: OperableType, divisor: OperableType) -> Expression:
    """
    Dividing operation, Error if the divisor is ``0``. To avoid the error, ``div0`` can be used instead.

    Parameters
    ----------
    dividend : OperableType
    divisor : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import div
    >>> m = Container()
    >>> a = Parameter(m, "a", records=210)
    >>> b = Parameter(m, "b")
    >>> b[...] = div(a, 3)
    >>> b.toValue()
    np.float64(70.0)

    """
    return expression.Expression(
        None, MathOp("div", (dividend, divisor)), None
    )


def div0(dividend: OperableType, divisor: OperableType) -> Expression:
    """
    Dividing operation, returns ``1e+299`` if the divisor is ``0``

    Parameters
    ----------
    dividend : OperableType
    divisor : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import div0
    >>> m = Container()
    >>> a = Parameter(m, "a", records=210)
    >>> b = Parameter(m, "b")
    >>> b[...] = div0(a, 0)
    >>> b.toValue()
    np.float64(1e+299)

    """
    return expression.Expression(
        None, MathOp("div0", (dividend, divisor)), None
    )


def dist(
    x1: OperableType,
    x2: OperableType,
) -> Expression:
    """
    Euclidean or L-2 Norm: ``sqrt(x1^2 + x2^2 + ... + xn^2)``

    Returns
    -------
    Expression

    Raises
    ------
    Exception
        In case both x1 and x2 are not a tuple or none.

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import dist
    >>> m = Container()
    >>> a = Parameter(m, "a", records=210)
    >>> b = Parameter(m, "b")
    >>> b[...] = dist(a, 100)

    """
    return expression.Expression(None, MathOp("eDist", (x1, x2)), None)


def factorial(x: int) -> Expression:
    """
    Factorial of ``x``: ``x!``

    Parameters
    ----------
    x : int

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import factorial
    >>> m = Container()
    >>> b = Parameter(m, "b")
    >>> b[...] = factorial(2)

    """
    return expression.Expression(None, MathOp("fact", (x,)), None)


def floor(x: OperableType) -> Expression:
    """
    The greatest integer less than or equal to ``x`` (i.e. ``floor(4.9)`` returns ``4``)

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import floor
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.9)
    >>> b = Parameter(m, "b")
    >>> b[...] = floor(a)
    >>> b.toValue()
    np.float64(3.0)

    """
    return expression.Expression(None, MathOp("floor", (x,)), None)


def fractional(x: OperableType) -> Expression:
    """
    Returns the fractional part of ``x`` (i.e. ``fractional(3.9)`` returns ``0.9``)

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import fractional
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.9)
    >>> b = Parameter(m, "b")
    >>> b[...] = fractional(a)
    >>> b.toValue()
    np.float64(0.8999999999999999)

    """
    return expression.Expression(None, MathOp("frac", (x,)), None)


def Min(*values) -> Expression:
    """
    Minimum value of the values, where the number of values may vary.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import Min
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", -2), ("i2", 0.3), ("i3", 2)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = Min(a[i], 1)
    >>> b.toList()
    [('i1', -2.0), ('i2', 0.3), ('i3', 1.0)]

    """
    return expression.Expression(None, MathOp("min", values), None)


def Max(*values) -> Expression:
    """
    Maximum value of the values, where the number of values may vary.

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import Max
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 2), ("i2", 0.3), ("i3", 2.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = Max(a[i], 1)
    >>> b.toList()
    [('i1', 2.0), ('i2', 1.0), ('i3', 2.5)]

    """
    return expression.Expression(None, MathOp("max", values), None)


def mod(x: OperableType, y: OperableType) -> Expression:
    """
    Remainder of ``x`` divided by ``y`` (i.e. ``mod(10, 3)`` returns ``1``)

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import mod
    >>> m = Container()
    >>> a = Parameter(m, "a", records=200)
    >>> b = Parameter(m, "b")
    >>> b[...] = mod(a, 3)
    >>> b.toValue()
    np.float64(2.0)

    """
    return expression.Expression(None, MathOp("mod", (x, y)), None)


def Round(x: OperableType, num_decimals: int = 0) -> Expression:
    """
    Round ``x`` to ``num_decimals`` decimal places (i.e. ``Round(3.14159, 2)`` returns ``3.14``)

    Parameters
    ----------
    x : OperableType
    num_decimals : int, optional

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import Round, div
    >>> m = Container()
    >>> a = Parameter(m, "a", records=200)
    >>> b = Parameter(m, "b")
    >>> b[...] = Round(div(a, 3), 2)
    >>> b.toValue()
    np.float64(66.67)

    """
    return expression.Expression(
        None, MathOp("round", (x, num_decimals)), None
    )


def sign(x: Symbol) -> Expression:
    """
    Sign of ``x`` returns ``1 if x > 0``, ``-1 if x < 0``, and ``0 if x = 0``

    Parameters
    ----------
    x : Symbol

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import sign
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 2), ("i2", -5.4), ("i3", 0)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = sign(a[i])
    >>> b.toList()
    [('i1', 1.0), ('i2', -1.0)]

    """
    return expression.Expression(None, MathOp("sign", (x,)), None)


def slexp(x: OperableType, S: int | float = 150) -> Expression:
    """
    Smooth (linear) exponential where ``S <= 150``. (Default ``S = 150``)

    Parameters
    ----------
    x : OperableType
    S : int | float, by default 150

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import slexp
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3)
    >>> b = Parameter(m, "b")
    >>> b[...] = slexp(a)
    >>> b.toValue()
    np.float64(20.085536923187668)

    """
    return expression.Expression(None, MathOp("slexp", (x, S)), None)


def sqexp(x: OperableType, S: int | float = 150) -> Expression:
    """
    Smooth (quadratic) exponential where ``S <= 150``. (Default ``S = 150``)

    Parameters
    ----------
    x : OperableType
    S : int | float, by default 150

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import sqexp
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3)
    >>> b = Parameter(m, "b")
    >>> b[...] = sqexp(a)
    >>> b.toValue()
    np.float64(20.085536923187668)

    """
    return expression.Expression(None, MathOp("sqexp", (x, S)), None)


def sqrt(x: OperableType, safe_cancel: bool = False) -> Expression:
    """
    Square root of ``x``

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import sqrt
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 54), ("i3", 0)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = sqrt(a[i])
    >>> b.toList()
    [('i1', 2.0), ('i2', 7.3484692283495345)]

    """
    return expression.Expression(
        None, MathOp("sqrt", (x,), safe_cancel=safe_cancel), None
    )


def truncate(x: OperableType) -> Expression:
    """
    Returns the integer part of ``x`` (i.e. ``truncate(3.9)`` returns ``3``)

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import truncate
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3.9)
    >>> b = Parameter(m, "b")
    >>> b[...] = truncate(a)
    >>> b.toValue()
    np.float64(3.0)

    """
    return expression.Expression(None, MathOp("trunc", (x,)), None)


def beta(x: OperableType, y: OperableType) -> Expression:
    """
    Beta function: ``B(x, y) = gamma(x) * gamma(y) / gamma(x + y) = (x-1)! * (y-1)! / (x + y - 1)!``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import beta
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3)
    >>> b = Parameter(m, "b")
    >>> b[...] = beta(a, 1)
    >>> b.toValue()
    np.float64(0.3333333333333333)

    """
    return expression.Expression(None, MathOp("beta", (x, y)), None)


def regularized_beta(
    x: int | float, y: int | float, z: int | float
) -> Expression:
    """
    Regularized Beta Function, See `MathWorld <https://mathworld.wolfram.com/RegularizedBetaFunction.html>`_

    Parameters
    ----------
    x : int | float
    y : int | float
    z : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import regularized_beta
    >>> m = Container()
    >>> a = Parameter(m, "a", records=3)
    >>> b = Parameter(m, "b")
    >>> b[...] = regularized_beta(0.5, a, 1)
    >>> b.toValue()
    np.float64(0.12500000000000003)

    """
    return expression.Expression(None, MathOp("betaReg", (x, y, z)), None)


def gamma(x: OperableType) -> Expression:
    """
    Gamma function: ``gamma(x) = (x-1)!``

    Parameters
    ----------
    x : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import gamma
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 7), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = gamma(a[i])
    >>> b.toList()
    [('i1', 6.0), ('i2', 720.0), ('i3', 1.772453850905516)]

    """
    return expression.Expression(None, MathOp("gamma", (x,)), None)


def regularized_gamma(x: int | float, a: int | float) -> Expression:
    """
    Lower Incomplete Regularized Gamma function, See `MathWorld <https://mathworld.wolfram.com/RegularizedGammaFunction.html>`_

    Parameters
    ----------
    x : int | float
    a : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import regularized_gamma
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 1), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = regularized_gamma(0.5, a[i])
    >>> b.toList()
    [('i1', 0.001751622556290824), ('i2', 0.3934693402873665), ('i3', 0.6826894921370857)]

    """
    return expression.Expression(None, MathOp("gammaReg", (x, a)), None)


def lse_max(*xs) -> Expression:
    """
    Smoothed Max via the Logarithm of the Sum of Exponentials: ``ln(exp(x1) + exp(x2) + ... + exp(xn))``

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import lse_max
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 10), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = lse_max(a[i], 5)
    >>> b.toList()
    [('i1', 5.313261687518223), ('i2', 10.006715348489118), ('i3', 5.011047744848594)]

    """
    if len(xs) < 1:
        raise ValidationError("lse_max requires at least 1 x")

    return expression.Expression(None, MathOp("lseMax", xs), None)


def lse_max_sc(t, *xs) -> Expression:
    """
    Scaled smoothed Max via the Logarithm of the Sum of Exponentials: ``lse_max_sc(T,x) = lse_max(Tx)/T``

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import lse_max_sc
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 100), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = lse_max_sc(7.5, a[i], 10.5)
    >>> b.toList()
    [('i1', 10.50000153604837), ('i2', 10.5), ('i3', 10.902826555965506)]

    """
    if len(xs) < 1:
        raise ValidationError("lse_max_sc requires at least 1 x")

    return expression.Expression(None, MathOp("lseMaxSc", xs + (t,)), None)


def lse_min(*xs) -> Expression:
    """
    Smoothed Min via the Logarithm of the Sum of Exponentials: ``-ln(exp(-x1) + exp(-x2) + ... + exp(-xn))``

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import lse_min
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 10), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = lse_min(a[i], 5)
    >>> b.toList()
    [('i1', 3.686738312481777), ('i2', 4.993284651510882), ('i3', 0.4889522551514062)]

    """
    if len(xs) < 1:
        raise ValidationError("lse_min requires at least 1 x")

    return expression.Expression(None, MathOp("lseMin", xs), None)


def lse_min_sc(t, *xs) -> Expression:
    """
    Scaled smoothed Min via the Logarithm of the Sum of Exponentials: ``lse_min_sc(T,x) = lse_min(Tx)/T``

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import lse_min_sc
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 100), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = lse_min_sc(7.5, a[i], 10.5)
    >>> b.toList()
    [('i1', 4.0), ('i2', 10.5), ('i3', 0.5)]

    """
    if len(xs) < 1:
        raise ValidationError("lse_min_sc requires at least 1 x")

    return expression.Expression(None, MathOp("lseMinSc", (t,) + xs), None)


def ncp_cm(x: Symbol, y: Symbol, z: float | int) -> Expression:
    """
    Chen-Mangasarian smoothing: ``x - z*ln(1 + exp((x-y)/z))``

    Parameters
    ----------
    x : Symbol
    y : Symbol
    z : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import ncp_cm
    >>> m = Container()
    >>> y = Parameter(m, "y", records=2)
    >>> b = Parameter(m, "b")
    >>> b[...] = ncp_cm(1, y, 0.5)
    >>> b.toValue()
    np.float64(0.9365359944785137)

    """
    return expression.Expression(None, MathOp("ncpCM", (x, y, z)), None)


def ncp_f(x: Symbol, y: Symbol, z: int | float = 0) -> Expression:
    """
    Fisher-Burmeister smoothing: ``sqrt(x^2 + y^2 + 2z) - x - y`` where ``z >= 0`` (default ``z = 0``)

    Parameters
    ----------
    x : Symbol
    y : Symbol
    z : int | float, optional

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import ncp_f
    >>> m = Container()
    >>> y = Parameter(m, "y", records=2)
    >>> b = Parameter(m, "b")
    >>> b[...] = ncp_f(1, y, 0.5)
    >>> b.toValue()
    np.float64(-0.5505102572168221)

    """
    return expression.Expression(None, MathOp("ncpF", (x, y, z)), None)


def ncpVUpow(
    r: Symbol,
    s: Symbol,
    mu: int | float = 0,
) -> Expression:
    """
    NCP Veelken-Ulbrich (smoothed min(r,s))

    Parameters
    ----------
    r : Symbol
    s : Symbol
    mu : int | float, optional

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import ncpVUpow
    >>> m = Container()
    >>> y = Parameter(m, "y", records=2)
    >>> b = Parameter(m, "b")
    >>> b[...] = ncpVUpow(1, y, 0.5)
    >>> b.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("ncpVUpow", (r, s, mu)), None)


def ncpVUsin(r: Symbol, s: Symbol, mu: int | float = 0) -> Expression:
    """
    NCP Veelken-Ulbrich (smoothed min(r,s))

    Parameters
    ----------
    r : Symbol
    s : Symbol
    mu : int | float, optional

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import ncpVUsin
    >>> m = Container()
    >>> y = Parameter(m, "y", records=2)
    >>> b = Parameter(m, "b")
    >>> b[...] = ncpVUsin(1, y, 0.5)
    >>> b.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("ncpVUsin", (r, s, mu)), None)


def poly(x, *args) -> Expression:
    """
    Polynomial function: ``p(x) = A[0] + A[1]*x + A[2]*x^2 + ... + A[n-1]*x^(n-1)``

    Returns
    -------
    Expression

    Raises
    ------
    ValidationError
        If the number of arguments (args) is less than 3 or if any of args is not an integer or a float.

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import poly
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", 10), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = poly(a[i], 15, 3, 4)
    >>> b.toList()
    [('i1', 91.0), ('i2', 445.0), ('i3', 17.5)]

    """
    if len(args) < 3:
        raise ValidationError("poly requires at least 3 arguments after x")

    return expression.Expression(None, MathOp("poly", (x,) + args), None)


def sigmoid(x: OperableType) -> Expression:
    """
    Sigmoid of ``x`` (i.e. ``1 / (1 + exp(-x))``)

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import sigmoid
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 4), ("i2", -1), ("i3", 0.5)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = sigmoid(a[i])
    >>> b.toList()
    [('i1', 0.9820137900379085), ('i2', 0.2689414213699951), ('i3', 0.6224593312018546)]

    """
    return expression.Expression(None, MathOp("sigmoid", (x,)), None)


def rand_binomial(n: int | float, p: int | float) -> Expression:
    """
    Generate a random number from the binomial distribution, where n is the
    number of trials and p the probability of success for each trial

    Parameters
    ----------
    n : int | float
    p : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import rand_binomial
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> p = Parameter(m, "p", domain=i, records=[("i1", 0.3), ("i2", 0.8), ("i3", 0.45)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = rand_binomial(75, p[i])
    >>> b.toList()
    [('i1', 21.0), ('i2', 63.0), ('i3', 25.0)]

    """
    return expression.Expression(None, MathOp("randBinomial", (n, p)), None)


def rand_linear(
    low: int | float, slope: int | float, high: int | float
) -> Expression:
    """
    Generate a random number between low and high with linear distribution.
    ``slope`` must be less than ``2 / (high - low)`` and greater than ``0``

    Parameters
    ----------
    low : int | float
    slope : int | float
    high : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import rand_linear
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> s = Parameter(m, "s", domain=i, records=[("i1", 0.03), ("i2", 0.008), ("i3", 0.04)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = rand_linear(75, s[i], 125)
    >>> b.toList()
    [('i1', 78.22119203430918), ('i2', 87.65662570307367), ('i3', 80.24583337516547)]

    """
    return expression.Expression(
        None, MathOp("randLinear", (low, slope, high)), None
    )


def rand_triangle(
    low: int | float, mid: int | float, high: int | float
) -> Expression:
    """
    Generate a random number between ``low`` and ``high`` with triangular distribution.
    ``mid`` is the most probable number.

    Parameters
    ----------
    low : int | float
    mid : int | float
    high : int | float

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import rand_triangle
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> s = Parameter(m, "s", domain=i, records=[("i1", 103), ("i2", 80), ("i3", 115)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = rand_triangle(75, s[i], 125)
    >>> b.toList()
    [('i1', 90.50632080153123), ('i2', 106.22102486822031), ('i3', 108.17756338250294)]

    """
    return expression.Expression(
        None, MathOp("randTriangle", (low, mid, high)), None
    )


def same_as(self: Set | Alias, other: Set | Alias | str) -> Expression:
    """
    Evaluates to true if this set is identical to the given set or alias, false otherwise.

    Parameters
    ----------
    other : Set | Alias

    Returns
    -------
    Expression

    Examples
    --------
    >>> import gamspy as gp
    >>> from gamspy.math import same_as
    >>> m = gp.Container()
    >>> i = gp.Set(m, name="i", records=["seattle", "san-diego"])
    >>> j = gp.Set(m, name="j", records=["new-york", "seattle"])
    >>> attr = gp.Parameter(m, "attr", domain = [i, j])
    >>> attr[i,j]  =  same_as(i, j)
    >>> attr.records.values.tolist()
    [['seattle', 'seattle', 1.0]]

    """
    return expression.Expression(None, MathOp("sameAs", (self, other)), None)


def slrec(x: OperableType, S: int | float = 1e-10) -> Expression:
    """
    Smooth (linear) reciprocal, where ``S >= 1e-10``. (Default ``S = 1e-10``)

    Parameters
    ----------
    x : OperableType
    S : int | float, by default 1e-10

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import slrec
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 1), ("i2", 0.8), ("i3", 15)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = slrec(a[i])
    >>> b.toList()
    [('i1', 1.0), ('i2', 1.25), ('i3', 0.06666666666666667)]

    """
    return expression.Expression(None, MathOp("slrec", (x, S)), None)


def sqrec(x: OperableType, S: int | float = 1e-10) -> Expression:
    """
    Smooth (quadratic) reciprocal, where ``S >= 1e-10``. (Default ``S = 1e-10``)

    Parameters
    ----------
    x : OperableType
    S : int | float, by default 1e-10

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import sqrec
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 1), ("i2", 0.8), ("i3", 15)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = sqrec(a[i])
    >>> b.toList()
    [('i1', 1.0), ('i2', 1.25), ('i3', 0.06666666666666667)]

    """
    return expression.Expression(None, MathOp("sqrec", (x, S)), None)


def entropy(x: OperableType) -> Expression:
    """
    Entropy function: ``-x*ln(x)`` where ``x >= 0``

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import entropy
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", 1), ("i2", 0.8), ("i3", 15)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = entropy(a[i])
    >>> b.toList()
    [('i2', 0.17851484105136778), ('i3', -40.62075301653315)]

    """
    return expression.Expression(None, MathOp("entropy", (x,)), None)


def errorf(x: OperableType) -> Expression:
    """
    Integral of the standard normal distribution from negative infinity to ``x``

    Parameters
    ----------
    x : int, float, Symbol

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Set, Parameter
    >>> from gamspy.math import errorf
    >>> m = Container()
    >>> i = Set(m, name="i", records=["i1", "i2", "i3"])
    >>> a = Parameter(m, "a", domain=i, records=[("i1", -2.5), ("i2", 0.8), ("i3", 1.7)])
    >>> b = Parameter(m, "b", domain=i)
    >>> b[i] = errorf(a[i])
    >>> b.toList()
    [('i1', 0.0062096653257761375), ('i2', 0.7881446014166034), ('i3', 0.955434537241457)]

    """
    return expression.Expression(None, MathOp("errorf", (x,)), None)


def ifthen(
    condition: Expression,
    yes_return: float | Expression,
    no_return: float | Expression,
) -> Expression:
    """
    If the logical condition is ``true``, the function returns ``yes_return``,
    else it returns ``no_return``

    Parameters
    ----------
    condition : Expression
    yes_return : float | Expression
    no_return : float | Expression

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy.math import ifthen
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> tt = gp.Parameter(m, "tt", records=2)
    >>> y = gp.Parameter(m, "y", records=2)
    >>> x = ifthen(tt == 2, 3, 4 + y)

    """
    condition._representation = utils._replace_equality_signs(
        condition.gamsRepr()
    )

    return expression.Expression(
        None, MathOp("ifthen", (condition, yes_return, no_return)), None
    )


def bool_and(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff both ``x and y`` are true

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import bool_and
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = bool_and(a > 10, b < 5)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("bool_and", (x, y)), None)


def bool_eqv(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``false`` iff exactly one argument is false

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import bool_eqv
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = bool_eqv(a > 10, b < 5)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("bool_eqv", (x, y)), None)


def bool_imp(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x is false`` or ``y is true``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import bool_imp
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = bool_imp(a < 10, b > 5)
    >>> c.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("bool_imp", (x, y)), None)


def bool_not(x: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x is false``

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import bool_not
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = bool_not(a > 10)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("bool_not", (x,)), None)


def bool_or(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x is true`` or ``y is true``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import bool_or
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = bool_or(a > 15, b < 5)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("bool_or", (x, y)), None)


def bool_xor(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff exactly one argument is ``false``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import bool_xor
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = bool_xor(a < 15, b > 5)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("bool_xor", (x, y)), None)


def rel_eq(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x == y``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import rel_eq
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = rel_eq(a, b)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("rel_eq", (x, y)), None)


def rel_ge(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x >= y``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import rel_ge
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = rel_ge(a, b)
    >>> c.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("rel_ge", (x, y)), None)


def rel_gt(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x > y``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import rel_gt
    >>> m = Container()
    >>> a = Parameter(m, "a", records=7)
    >>> b = Parameter(m, "b", records=7)
    >>> c = Parameter(m, "c")
    >>> c[...] = rel_gt(a, b)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("rel_gt", (x, y)), None)


def rel_le(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x <= y``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import rel_le
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=11)
    >>> c = Parameter(m, "c")
    >>> c[...] = rel_le(a, b)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("rel_le", (x, y)), None)


def rel_lt(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x < y``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import rel_lt
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=17)
    >>> c = Parameter(m, "c")
    >>> c[...] = rel_lt(a, b)
    >>> c.toValue()
    np.float64(1.0)

    """
    return expression.Expression(None, MathOp("rel_lt", (x, y)), None)


def rel_ne(x: OperableType, y: OperableType) -> Expression:
    """
    Returns ``true`` iff ``x != y``

    Parameters
    ----------
    x : OperableType
    y : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import rel_ne
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b", records=12)
    >>> c = Parameter(m, "c")
    >>> c[...] = rel_ne(a, b)
    >>> c.toValue()
    np.float64(0.0)

    """
    return expression.Expression(None, MathOp("rel_ne", (x, y)), None)


def map_value(x: OperableType) -> Expression:
    """
    Returns an integer value that indicates what special value (if any) is stored in the input.
    Possible results:
    0: is not a special value
    4: is UNDF (undefined)
    5: is NA (not available)
    6: is INF
    7: is -INF
    8: is EPS

    Parameters
    ----------
    x : OperableType

    Returns
    -------
    Expression

    Examples
    --------
    >>> from gamspy import Container, Parameter
    >>> from gamspy.math import map_value
    >>> m = Container()
    >>> a = Parameter(m, "a", records=12)
    >>> b = Parameter(m, "b")
    >>> b[...] = map_value(a)
    >>> b.toValue()
    np.float64(0.0)
    >>> a[...] = float('inf')
    >>> b[...] = map_value(a)
    >>> b.toValue()
    np.float64(6.0)

    """
    return expression.Expression(None, MathOp("mapval", (x,)), None)



from __future__ import annotations

import os
from typing import Any

from gamspy.exceptions import ValidationError

configuration: dict[str, Any] = dict()


def _set_default_options() -> None:
    # Set the default for GAMS_SYSDIR
    try:
        import gamspy_base

        configuration["GAMS_SYSDIR"] = gamspy_base.directory
    except ModuleNotFoundError:
        ...

    sysdir = os.getenv("GAMSPY_GAMS_SYSDIR", None)
    if sysdir is not None:
        configuration["GAMS_SYSDIR"] = sysdir

    # Enable all validations by default
    validate = int(os.getenv("GAMSPY_VALIDATION", 1))
    configuration["VALIDATION"] = validate

    # Check for domain violation by default
    validate = int(os.getenv("GAMSPY_DOMAIN_VALIDATION", 1))
    configuration["DOMAIN_VALIDATION"] = validate

    # Solver option validation. Enabled by default.
    validate_solver_options = int(
        os.getenv("GAMSPY_SOLVER_OPTION_VALIDATION", 1)
    )
    configuration["SOLVER_OPTION_VALIDATION"] = validate_solver_options

    # Special value mapping
    map_special_values = int(os.getenv("GAMSPY_MAP_SPECIAL_VALUES", 1))
    configuration["MAP_SPECIAL_VALUES"] = map_special_values

    # Lazy evaluation
    evaluate_lazily = int(os.getenv("GAMSPY_LAZY_EVALUATION", 0))
    configuration["LAZY_EVALUATION"] = evaluate_lazily


def set_options(options: dict[str, Any]) -> None:
    """
    Sets the given configuration options.

    Parameters
    ----------
    options : dict[str, Any]

    Raises
    ------
    ValidationError
        In case the given options are not in dict type.

    Examples
    --------
    >>> import gamspy as gp
    >>> gp.set_options({"DOMAIN_VALIDATION": 1})

    """
    if not isinstance(options, dict):
        raise ValidationError(
            f"`options` must be a dictionary but found: `{type(options)}`"
        )

    for key, value in options.items():
        configuration[key] = value


def get_option(name: str) -> Any:
    """
    Returns the requested option.

    Parameters
    ----------
    name : str
        Option name.

    Returns
    -------
    Any
        The value of the option.

    Raises
    ------
    KeyError
        In case the option is not set.

    Examples
    --------
    >>> import gamspy as gp
    >>> gp.set_options({"DOMAIN_VALIDATION": 0})
    >>> gp.get_option("DOMAIN_VALIDATION")
    0
    >>> gp.set_options({"DOMAIN_VALIDATION": 1})

    """
    return configuration[name]



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._symbols.alias as alias
import gamspy._symbols.implicits as implicits
import gamspy._symbols.set as gams_set
from gamspy._symbols.implicits.implicit_symbol import ImplicitSymbol

if TYPE_CHECKING:
    import pandas as pd

    from gamspy import Equation, Set
    from gamspy._algebra.expression import Expression


class ImplicitEquation(ImplicitSymbol):
    def __init__(
        self,
        parent: Equation,
        name: str,
        type: str,
        domain: list[Set | str],
    ) -> None:
        """Implicit Equation

        Parameters
        ----------
        parent : Equation
        name : str
        domain : List[Set | str]
        """
        super().__init__(parent, name, domain)
        self.type = type

    def __repr__(self) -> str:
        return f"ImplicitEquation(parent={self.parent}, name='{self.name}', domain={self.domain}, type={self.type})"

    @property
    def l(self):  # noqa: E741,E743
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.l.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @l.setter
    def l(self, value: int | float | Expression):
        # b[t].l = 30 -> b.l[t] = 30
        self.parent.l[self.domain] = value

    @property
    def m(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.m.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @m.setter
    def m(self, value: int | float | Expression):
        # b[t].m = 30 -> b.m[t] = 30
        self.parent.m[self.domain] = value

    @property
    def lo(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.lo.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @lo.setter
    def lo(self, value: int | float | Expression):
        # b[t].lo = 30 -> b.lo[t] = 30
        self.parent.lo[self.domain] = value

    @property
    def up(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.up.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @up.setter
    def up(self, value: int | float | Expression):
        # b[t].up = 30 -> b.up[t] = 30
        self.parent.up[self.domain] = value

    @property
    def scale(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.scale.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @scale.setter
    def scale(self, value: int | float | Expression):
        # b[t].scale = 30 -> b.scale[t] = 30
        self.parent.scale[self.domain] = value

    @property
    def stage(self):
        return implicits.ImplicitParameter(
            self.parent,
            name=self.parent.stage.name,
            domain=self.domain,
            scalar_domains=self._scalar_domains,
        )

    @stage.setter
    def stage(self, value: int | float | Expression):
        # b[t].stage = 30 -> b.stage[t] = 30
        self.parent.stage[self.domain] = value

    @property
    def range(self):
        return self.parent.range

    @property
    def slacklo(self):
        return self.parent.slacklo

    @property
    def slackup(self):
        return self.parent.slackup

    @property
    def slack(self):
        return self.parent.slack

    @property
    def records(self) -> pd.DataFrame | float | None:
        if self.parent.records is None:
            return None

        recs = self.parent.records
        for idx, literal in self._scalar_domains:
            column_name = recs.columns[idx]
            recs = recs[recs[column_name] == literal]

        return recs

    def gamsRepr(self) -> str:
        representation = f"{self.name}"
        domain = list(self.domain)
        for i, d in self._scalar_domains:
            domain.insert(i, d)

        if len(domain):
            set_strs = []
            for set in domain:
                if isinstance(
                    set, (gams_set.Set, alias.Alias, implicits.ImplicitSet)
                ):
                    set_strs.append(set.gamsRepr())
                elif isinstance(set, str):
                    if set == "*":
                        set_strs.append(set)
                    else:
                        set_strs.append('"' + set + '"')

            domain_str = "(" + ",".join(set_strs) + ")"

            representation += domain_str

        return representation



from __future__ import annotations
import importlib
import shutil

import sys
from typing import Annotated, Iterable, Optional, Union

import certifi
import typer
from gamspy.exceptions import GamspyException, ValidationError
import gamspy.utils as utils
import os
import subprocess
from .util import add_solver_entry

app = typer.Typer(
    rich_markup_mode="rich",
    short_help="To install licenses and solvers.",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy install license <access_code> or <path/to/license/file> | gamspy install solver <solver_name>",
    context_settings={"help_option_names": ["-h", "--help"]},
)

@app.command(
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy install license <access_code> or <path/to/license/file>",
    short_help="To install a new license"
)
def license(
    license: Annotated[str, typer.Argument(help="access code or path to the license file.")],
    checkout_duration: Optional[int] = typer.Option(None, "--checkout-duration", "-c", help="Specify a duration in hours to checkout a session."),
    renew: Optional[str] = typer.Option(None, "--renew", "-r", help="Specify a file path to a license file to extend a session."),
    output: Optional[str] = typer.Option(None, "--output", "-o", help="Specify a file path to write the license file."),
    uses_port: Annotated[Union[int, None], typer.Option("--uses-port", help="Interprocess communication starting port.")] = None,
):
    import json
    from urllib.parse import urlencode

    import urllib3

    os.makedirs(utils.DEFAULT_DIR, exist_ok=True)

    is_alp = not os.path.isfile(license)

    if is_alp and len(license) != 36:
        raise ValidationError(
            f"Access code is a 36 character string or an absolute path to the "
            f"license file but {len(license)} character string ({license}) provided."
        )

    gamspy_base_dir = utils._get_gamspy_base_directory()
    license_path = os.path.join(utils.DEFAULT_DIR, "gamspy_license.txt")

    if is_alp:
        alp_id = license
        http = urllib3.PoolManager(
            cert_reqs="CERT_REQUIRED",
            ca_certs=certifi.where()
        )
        encoded_args = urlencode({"access_token": alp_id})
        request = http.request(
            "GET", "https://license.gams.com/license-type?" + encoded_args
        )
        if request.status != 200:
            raise ValidationError(
                f"License server did not respond in an expected way. Request status: {request.status}. Reason: {request.data.decode('utf-8', errors='replace')}"
            )

        data = request.data.decode("utf-8", errors="replace")
        cmex_type = json.loads(data)["cmex_type"]
        if cmex_type not in ("gamspy", "gamspy++", "gamsall"):
            raise ValidationError(
                f"Given access code `{alp_id} ({cmex_type})` is not valid for GAMSPy. "
                "Make sure that you use a GAMSPy license, not a GAMS license."
            )

        command = [os.path.join(gamspy_base_dir, "gamsgetkey"), alp_id]

        if uses_port:
            command.append("-u")
            command.append(str(uses_port))

        if checkout_duration:
            command.append("-c")
            command.append(str(checkout_duration))

        if checkout_duration:
            command.append("-r")
            command.append(str(renew))
        
        if output:
            command.append("-o")
            command.append(output)

        process = subprocess.run(
            command,
            text=True,
            capture_output=True,
        )
        if process.returncode:
            raise ValidationError(process.stderr)

        if output:
            with open(output) as file:
                license_text = file.read()
        else:
            license_text = process.stdout
            
        lines = license_text.splitlines()
        license_type = lines[0][54]
        if license_type == "+":
            if lines[2][:2] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given access code `{alp_id}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )
        else:
            if lines[2][8:10] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given access code `{alp_id}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )

        with open(license_path, "w", encoding="utf-8") as file:
            file.write(license_text)
    else:
        with open(license) as file:
            lines = file.read().splitlines()

        license_type = lines[0][54]
        if license_type == "+":
            if lines[2][:2] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given license file `{license}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )
        else:
            if lines[2][8:10] not in ("00", "07", "08", "09"):
                raise ValidationError(
                    f"Given license file `{license}` is not valid for GAMSPy. "
                    "Make sure that you use a GAMSPy license, not a GAMS license."
                )

        shutil.copy(license, license_path)

def append_dist_info(files, gamspy_base_dir: str):
    """Updates dist-info/RECORD in site-packages for pip uninstall"""
    import gamspy as gp

    gamspy_path: str = gp.__path__[0]
    dist_info_path = f"{gamspy_path}-{gp.__version__}.dist-info"

    with open(
        dist_info_path + os.sep + "RECORD", "a", encoding="utf-8"
    ) as record:
        gamspy_base_relative_path = os.sep.join(
            gamspy_base_dir.split(os.sep)[-3:]
        )

        lines = []
        for file in files:
            line = f"{gamspy_base_relative_path}{os.sep}{file},,"
            lines.append(line)

        record.write("\n".join(lines) + "\n")

@app.command(
    short_help="To install solvers",
    help="[bold][yellow]Examples[/yellow][/bold]: gamspy install solver <solver_name>"
)
def solver(
    solver: list[str] = typer.Argument(
        None,
        help="solver names to be installed",
        autocompletion=lambda: [s.lower() for s in utils.getAvailableSolvers()]
    ),
    install_all_solvers: bool = typer.Option(
        False,
        "--install-all-solvers",
        help="Installs all available add-on solvers."
    ),
    existing_solvers: bool = typer.Option(
        False,
        "--existing-solvers",
        help="Reinstalls previously installed add-on solvers."
    ),
    skip_pip_install: bool = typer.Option(
        False,
        "--skip-pip-install",
        "-s",
        help="If you already have the solver installed, skip pip install and update gamspy installed solver list."
    ),
    use_uv: bool = typer.Option(
        False,
        "--use-uv",
        help="Use uv instead of pip to install solvers."
    ),
):
    try:
        import gamspy_base
    except ModuleNotFoundError as e:
        e.msg = "You must first install gamspy_base to use this functionality"
        raise e

    addons_path = os.path.join(utils.DEFAULT_DIR, "solvers.txt")
    os.makedirs(utils.DEFAULT_DIR, exist_ok=True)

    def install_addons(addons: Iterable[str]):
        for item in addons:
            solver_name = item.lower()

            if solver_name.upper() in gamspy_base.default_solvers:
                print(f"`{solver_name}` is a default solver, skipping...")
                continue
            installable_solvers = utils.getInstallableSolvers(gamspy_base.directory)
            if solver_name.upper() not in installable_solvers:
                raise ValidationError(
                    f'Given solver name ("{solver_name}") is not valid. Available'
                    f" solvers that can be installed: {installable_solvers}")
            
            installed_solvers = utils.getInstalledSolvers(gamspy_base.directory)
            if solver_name.upper() in installed_solvers:
                print(f"`{solver_name}` is already installed, skipping...")
                continue

            if not skip_pip_install:

                solver_version = gamspy_base.__version__
                # install specified solver
                if use_uv:
                    command = [
                        "uv",
                        "pip",
                        "--python-preference",
                        "only-system",
                        "install",
                        f"gamspy-{solver_name}=={solver_version}",
                        "--force-reinstall",
                    ]
                else:
                    command = [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        f"gamspy-{solver_name}=={solver_version}",
                        "--force-reinstall",
                    ]
                try:
                    _ = subprocess.run(command, check=True, stderr=subprocess.PIPE)
                except subprocess.CalledProcessError as e:
                    raise GamspyException(
                        f"Could not install gamspy-{solver_name}. Please check your internet connection. If it's not related to your internet connection, PyPI servers might be down. Please retry it later. Here is the error message of pip:\n\n{e.stderr.decode('utf-8')}"
                    ) from e
            else:
                try:
                    solver_lib = importlib.import_module(
                        f"gamspy_{solver_name}"
                    )
                except ModuleNotFoundError as e:
                    e.msg = f"You must install gamspy-{solver_name} first!"
                    raise e

            # copy solver files to gamspy_base
            gamspy_base_dir = utils._get_gamspy_base_directory()
            solver_lib = importlib.import_module(f"gamspy_{solver_name}")

            file_paths = solver_lib.file_paths
            for file in file_paths:
                shutil.copy(file, gamspy_base_dir)

            files = solver_lib.files
            verbatims = [solver_lib.verbatim]
            append_dist_info(files, gamspy_base_dir)
            add_solver_entry(gamspy_base_dir, solver_name, verbatims)

            try:
                with open(addons_path) as file:
                    installed = file.read().splitlines()
                    installed = [
                        solver
                        for solver in installed
                        if solver != "" and solver != "\n"
                    ]
            except FileNotFoundError:
                installed = []

            with open(addons_path, "w") as file:
                if solver_name.upper() not in installed:
                    file.write("\n".join(installed + [solver_name.upper()]))

    if install_all_solvers:
        available_solvers = utils.getAvailableSolvers()
        installed_solvers = utils.getInstalledSolvers(gamspy_base.directory)
        diff = []
        for available_solver in available_solvers:
            if available_solver not in installed_solvers:
                diff.append(available_solver)

        install_addons(diff)
        return

    if existing_solvers:
        try:
            with open(addons_path) as file:
                solvers = file.read().splitlines()
                solvers = [solver for solver in solvers if solver != "" and solver != "\n"]
                install_addons(solvers)
                return
        except FileNotFoundError as e:
            raise ValidationError("No existing add-on solvers found!") from e

    if solver is None:
        raise ValidationError(
            "Solver name is missing: `gamspy install solver <solver_name>`"
        )

    install_addons(solver)

if __name__ == "__main__":
    app()



# flake8: noqa
from __future__ import annotations

from gamspy._algebra.domain import Domain
from gamspy._algebra.number import Number
from gamspy._algebra.operation import Card
from gamspy._algebra.operation import Ord
from gamspy._algebra.operation import Product
from gamspy._algebra.operation import Smax
from gamspy._algebra.operation import Smin
from gamspy._algebra.operation import Sum
from gamspy._algebra.operation import Sand
from gamspy._algebra.operation import Sor



from __future__ import annotations

import atexit
import os
import platform
import signal
import socket
import subprocess
import tempfile
import threading
import time
import traceback
import weakref
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING

import gams.transfer as gt

import gamspy as gp
import gamspy._miro as miro
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._backend.backend import backend_factory
from gamspy._config import get_option
from gamspy._extrinsic import ExtrinsicLibrary
from gamspy._miro import MiroJSONEncoder
from gamspy._model import Problem, Sense
from gamspy._options import Options
from gamspy._workspace import Workspace
from gamspy.exceptions import FatalError, GamspyException, ValidationError

if TYPE_CHECKING:
    import io
    from typing import Any, Literal

    from pandas import DataFrame

    from gamspy import (
        Alias,
        Equation,
        EquationType,
        Model,
        Parameter,
        Set,
        Variable,
    )
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation

LOOPBACK = "127.0.0.1"
IS_MIRO_INIT = os.getenv("MIRO", False)
MIRO_GDX_IN = os.getenv("GAMS_IDC_GDX_INPUT", None)
MIRO_GDX_OUT = os.getenv("GAMS_IDC_GDX_OUTPUT", None)
is_windows = platform.system() == "Windows"


def add_sysdir_to_path(system_directory: str) -> None:
    if is_windows:
        if "PATH" in os.environ:
            if not os.environ["PATH"].startswith(
                system_directory + os.pathsep
            ):
                os.environ["PATH"] = (
                    system_directory + os.pathsep + os.environ["PATH"]
                )
        else:
            os.environ["PATH"] = system_directory


def open_connection(
    system_directory: str, process_directory: str, license_path: str
) -> tuple[socket.socket, subprocess.Popen]:
    TIMEOUT = 30

    initial_pf_file = os.path.join(process_directory, "gamspy.pf")
    with open(initial_pf_file, "w") as file:
        file.write(
            'incrementalMode="2"\n'
            f'procdir="{process_directory}"\n'
            f'license="{license_path}"\n'
            f'curdir="{os.getcwd()}"\n'
        )

    process = subprocess.Popen(
        [
            os.path.join(system_directory, "gams"),
            "GAMSPY_JOB",
            "pf",
            initial_pf_file,
        ],
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        errors="replace",
        start_new_session=platform.system() != "Windows",
    )

    port_info = process.stdout.readline().strip()

    try:
        port = int(port_info.removeprefix("port: "))
    except ValueError as e:
        raise ValidationError(
            f"Error while reading the port! {port_info + process.stdout.read()}"
        ) from e

    def handler(signum, frame):
        if platform.system() != "Windows":
            os.kill(process.pid, signal.SIGINT)

    if threading.current_thread() is threading.main_thread():
        signal.signal(signal.SIGINT, handler)

    start = time.time()
    while True:
        if process.poll() is not None:  # pragma: no cover
            raise ValidationError(process.communicate()[0])

        try:
            new_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            new_socket.connect((LOOPBACK, port))
            break
        except (ConnectionRefusedError, OSError) as e:
            new_socket.close()
            end = time.time()

            if end - start > TIMEOUT:  # pragma: no cover
                raise FatalError(
                    f"Timeout while establishing the connection with socket. {process.communicate()[0]}"
                ) from e

    return new_socket, process


def get_system_directory(system_directory: str | None) -> str:
    if system_directory is not None:
        return system_directory

    return get_option("GAMS_SYSDIR")


def check_response(response: bytes, job_name: str) -> None:
    GAMS_STATUS = {
        1: "Solver is to be called, the system should never return this number.",
        2: "There was a compilation error.",
        3: "There was an execution error.",
        4: "System limits were reached.",
        5: "There was a file error.",
        6: "There was a parameter error.",
        7: "The solve has failed due to a license error. The license you are using may impose model size limits (demo/community license) or you are using a GAMSPy incompatible professional license. Please contact sales@gams.com to find out about license options.",
        8: "There was a GAMS system error.",
        9: "GAMS could not be started.",
        10: "Out of memory.",
        11: "Out of disk.",
        109: "Could not create process/scratch directory.",
        110: "Too many process/scratch directories.",
        112: "Could not delete the process/scratch directory.",
        113: "Could not write the script gamsnext.",
        114: "Could not write the parameter file.",
        115: "Could not read environment variable.",
        400: "Could not spawn the GAMS language compiler (gamscmex).",
        401: "Current directory (curdir) does not exist.",
        402: "Cannot set current directory (curdir).",
        404: "Blank in system directory (UNIX only).",
        405: "Blank in current directory (UNIX only).",
        406: "Blank in scratch extension (scrext)",
        407: "Unexpected cmexRC.",
        408: "Could not find the process directory (procdir).",
        409: "CMEX library not be found (experimental).",
        410: "Entry point in CMEX library could not be found (experimental).",
        411: "Blank in process directory (UNIX only).",
        412: "Blank in scratch directory (UNIX only).",
        909: "Cannot add path / unknown UNIX environment / cannot set environment variable.",
        1000: "Driver error: incorrect command line parameters for gams.",
        2000: "Driver error: internal error: cannot install interrupt handler.",
        3000: "Driver error: problems getting current directory.",
        4000: "Driver error: internal error: GAMS compile and execute module not found.",
        5000: "Driver error: internal error: cannot load option handling library.",
    }

    value = response[: response.find(b"#")].decode("ascii")
    if not value:
        raise FatalError(
            "Error while getting the return code from GAMS backend. This means that GAMS is in a bad state. Try to backtrack for previous errors."
        )

    return_code = int(value)

    if return_code in GAMS_STATUS:
        try:
            info = GAMS_STATUS[return_code]
        except IndexError:  # pragma: no cover
            info = ""

        raise GamspyException(
            f"Return code {return_code}. {info} Check {job_name + '.lst'} for more information.",
            return_code,
        )


class Container(gt.Container):
    """
    A container is an object that holds all symbols and operates on them.

    Parameters
    ----------
    load_from : str, Container, gt.Container, optional
        Path to the GDX file to be loaded from, by default None
    system_directory : str, optional
        Path to the directory that holds the GAMS installation, by default None
    working_directory : str, optional
        Path to the working directory to store temporary files such .lst, .gms,
        .gdx, .g00 files.
    debugging_level : str, optional
        Decides on keeping the temporary files generate by GAMS, by default
        "keep_on_error"
    options : Options, optional
        Global options for the overall execution

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i")

    """

    def __init__(
        self,
        load_from: str | Container | gt.Container | None = None,
        system_directory: str | None = None,
        working_directory: str | None = None,
        debugging_level: Literal[
            "keep", "keep_on_error", "delete"
        ] = "keep_on_error",
        options: Options | None = None,
        output: io.TextIOWrapper | None = None,
    ):
        self.output = output
        self._gams_string = ""
        self.models: dict[str, Model] = dict()
        if IS_MIRO_INIT:
            atexit.register(self._write_miro_files)

        self._is_socket_open = True

        system_directory = get_system_directory(system_directory)
        add_sysdir_to_path(system_directory)

        self._unsaved_statements: list = []

        super().__init__(system_directory=system_directory)
        self._license_path = utils._get_license_path(self.system_directory)
        self._network_license = self._is_network_license()

        self._debugging_level = debugging_level
        self._workspace = Workspace(debugging_level, working_directory)
        self._working_directory = self._workspace.working_directory
        if len(self._working_directory) > 204:
            raise ValidationError(
                f"Working directory path length cannot be greater than 204! Given working directory path `{self._working_directory}` has length of `{len(self._working_directory)}`"
            )

        self._process_directory = tempfile.mkdtemp(dir=self.working_directory)

        self._job, self._gdx_in, self._gdx_out = self._setup_paths()

        self._temp_container = gt.Container(
            system_directory=self.system_directory
        )

        self._options = validation.validate_global_options(options)

        # needed for miro
        self._miro_input_symbols: list[str] = []
        self._miro_output_symbols: list[str] = []

        self._socket, self._process = open_connection(
            self.system_directory, self._process_directory, self._license_path
        )
        weakref.finalize(
            self,
            self._release_resources,
            self._is_socket_open,
            self._socket,
            self._process,
        )

        self._is_restarted = False
        if load_from is not None:
            if not isinstance(load_from, (str, gt.Container)):
                raise ValidationError(
                    f"`load_from` must be of type str or Container but found {type(load_from)}"
                )

            if isinstance(load_from, str) and load_from[-4:] not in (
                ".gdx",
                ".g00",
            ):
                raise ValidationError(
                    f"`load_from` must end with .gdx or .g00 but found {load_from}"
                )

            if isinstance(load_from, str) and load_from.endswith(".g00"):
                self._options._set_debug_options(
                    {"restart": load_from, "gdxSymbols": "all"}
                )
                self._synch_with_gams(gams_to_gamspy=True)
                self._options._set_debug_options(dict())
                self._clean_modified_symbols()
                self._unsaved_statements = []
                self._is_restarted = True
            else:
                self._read(load_from)
                if not isinstance(load_from, gt.Container):
                    self._unsaved_statements = []
                    self._clean_modified_symbols()
                    self._add_statement(f"$declareAndLoad {load_from}")

                self._synch_with_gams()

    def __enter__(self):
        pid = os.getpid()
        tid = threading.get_native_id()
        gp._ctx_managers[(pid, tid)] = self

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pid = os.getpid()
        tid = threading.get_native_id()

        try:
            del gp._ctx_managers[(pid, tid)]
        except KeyError:
            ...

    def __repr__(self) -> str:
        return f"Container(system_directory='{self.system_directory}', working_directory='{self.working_directory}', debugging_level='{self._debugging_level}')"

    def __str__(self):
        if len(self):
            return f"<Container ({hex(id(self))}) with {len(self)} symbols: {self.data.keys()}>"

        return f"<Empty Container ({hex(id(self))})>"

    @property
    def working_directory(self) -> str:
        """
        Working directory path.

        Returns
        -------
        str
        """
        return self._working_directory

    @property
    def in_miro(self) -> bool:
        """
        When running a GAMSPy job from GAMS MIRO, you may not want to
        perform certain expensive operations, such as loading MIRO input
        data from an Excel workbook, as this data comes from MIRO. In that
        case, one can conditionally load the data by using the ``in_miro``
        attribute of `Container`.

        Returns
        -------
        bool
        """
        return MIRO_GDX_IN is not None

    def _is_network_license(self) -> bool:
        with open(self._license_path, encoding="utf-8") as file:
            lines = file.readlines()

        return bool("+" in lines[0] and lines[4][47] == "N")

    @staticmethod
    def _release_resources(
        is_socket_open: bool, socket: socket.socket, process: subprocess.Popen
    ):
        if is_socket_open:
            try:
                socket.sendall(b"stop")
                socket.close()

                if not process.stdout.closed:
                    process.stdout.close()

                while process.poll() is None:
                    ...
            except OSError:  # Socket may have been already closed.
                ...

    def _read_output(self, output: io.TextIOWrapper | None) -> None:
        if output is not None:
            while True:
                data = self._process.stdout.readline()
                output.write(data)
                output.flush()
                if data.startswith("--- Job ") and "elapsed" in data:
                    break

    def _interrupt(self) -> None:
        if platform.system() == "Windows":
            os.kill(self._process.pid, signal.CTRL_C_EVENT)
        else:
            os.kill(self._process.pid, signal.SIGINT)

    def _send_job(
        self,
        job_name: str,
        pf_file: str,
        output: io.TextIOWrapper | None = None,
    ):
        try:
            # Send pf file
            self._socket.sendall(pf_file.encode("utf-8"))

            # Read output
            self._read_output(output)

            # Receive response
            response = self._socket.recv(256)
            check_response(response, job_name)
        except ConnectionError as e:
            raise FatalError(
                f"There was an error while communicating with GAMS server: {e}",
            ) from e

    def _write_miro_files(self):
        # create conf_<model>/<model>_io.json
        try:
            encoder = MiroJSONEncoder(self)
            encoder.write_json()
        except Exception:
            self._release_resources(
                self._is_socket_open, self._socket, self._process
            )
            self._is_socket_open = False
            traceback.print_exc()
            os._exit(1)

    def _add_statement(self, statement) -> None:
        self._unsaved_statements.append(statement)

    def _cast_symbols(self, symbol_names: list[str] | None = None) -> None:
        """Casts GTP symbols to GAMSPy symbols"""
        symbol_names = symbol_names if symbol_names else list(self.data.keys())

        for symbol_name in symbol_names:
            gtp_symbol = self.data[symbol_name]
            new_domain = [
                (self.data[member.name] if type(member) is not str else member)
                for member in gtp_symbol.domain
            ]

            del self.data[symbol_name]

            if isinstance(gtp_symbol, gt.Alias):
                alias_with = self.data[gtp_symbol.alias_with.name]
                _ = gp.Alias._constructor_bypass(
                    self, gtp_symbol._name, alias_with
                )
            elif isinstance(gtp_symbol, gt.UniverseAlias):
                _ = gp.UniverseAlias._constructor_bypass(
                    self,
                    gtp_symbol._name,
                )
            elif isinstance(gtp_symbol, gt.Set):
                _ = gp.Set._constructor_bypass(
                    self,
                    gtp_symbol._name,
                    new_domain,
                    gtp_symbol._is_singleton,
                    gtp_symbol._records,
                    gtp_symbol._description,
                )
            elif isinstance(gtp_symbol, gt.Parameter):
                _ = gp.Parameter._constructor_bypass(
                    self,
                    gtp_symbol._name,
                    new_domain,
                    gtp_symbol._records,
                    gtp_symbol._description,
                )
            elif isinstance(gtp_symbol, gt.Variable):
                _ = gp.Variable._constructor_bypass(
                    self,
                    gtp_symbol._name,
                    gtp_symbol._type,
                    new_domain,
                    gtp_symbol._records,
                    gtp_symbol._description,
                )
            elif isinstance(gtp_symbol, gt.Equation):
                symbol_type = gtp_symbol.type
                if gtp_symbol.type in ("eq", "leq", "geq"):
                    symbol_type = "regular"

                _ = gp.Equation._constructor_bypass(
                    self,
                    gtp_symbol._name,
                    symbol_type,
                    new_domain,
                    gtp_symbol._records,
                    gtp_symbol._description,
                )

    def _delete_autogenerated_symbols(self) -> None:
        """
        Removes autogenerated model attributes, objective variable and equation from
        the container
        """
        autogenerated_symbol_names = self._get_autogenerated_symbol_names()

        for name in autogenerated_symbol_names:
            if name in self.data:
                del self.data[name]

    def _setup_paths(self) -> tuple[str, str, str]:
        suffix = "_" + utils._get_unique_name()
        job = os.path.join(self.working_directory, suffix)
        gdx_in = os.path.join(self.working_directory, f"{suffix}in.gdx")
        gdx_out = os.path.join(self.working_directory, f"{suffix}out.gdx")

        return job, gdx_in, gdx_out

    def _get_autogenerated_symbol_names(self) -> list[str]:
        names = []
        for name in self.data:
            if name.startswith(gp.Model._generate_prefix):
                names.append(name)

        return names

    def _get_modified_symbols(self) -> list[str]:
        modified_names = []

        for name, symbol in self:
            if symbol.modified:
                if (
                    type(symbol) is gp.Alias
                    and symbol.alias_with.name not in modified_names  # type: ignore
                ):
                    modified_names.append(symbol.alias_with.name)  # type: ignore

                modified_names.append(name)

        return modified_names

    def _clean_modified_symbols(self) -> None:
        for symbol in self.data.values():
            if type(symbol) is gp.Alias:
                symbol.alias_with.modified = False  # type: ignore

            symbol.modified = False

    def _synch_with_gams(
        self,
        relaxed_domain_mapping: bool = False,
        gams_to_gamspy: bool = False,
    ) -> DataFrame | None:
        runner = backend_factory(self, self._options, output=self.output)
        summary = runner.run(relaxed_domain_mapping, gams_to_gamspy)

        if self._options and self._options.seed is not None:
            # Required for correct seeding. Seed can only be set in the first run.
            self._options.seed = None

        if IS_MIRO_INIT:
            miro._write_default_gdx_miro(self)

        return summary

    def _generate_gams_string(
        self,
        gdx_in: str,
        modified_names: list[str],
    ) -> str:
        LOADABLE = (gp.Set, gp.Parameter, gp.Variable, gp.Equation)
        MIRO_INPUT_TYPES = (gp.Set, gp.Parameter)

        strings = ["$onMultiR", "$onUNDF"]
        for statement in self._unsaved_statements:
            if type(statement) is str:
                strings.append(statement)
            else:
                strings.append(statement.getDeclaration())

        if modified_names:
            loadables = []
            for name in modified_names:
                symbol = self[name]
                if (
                    type(symbol) in LOADABLE
                    and not name.startswith(gp.Model._generate_prefix)
                    and symbol.synchronize
                ):
                    loadables.append(symbol)

            if loadables:
                strings.append(f"$gdxIn {gdx_in}")
                for loadable in loadables:
                    if (
                        type(loadable) in MIRO_INPUT_TYPES
                        and loadable._is_miro_input
                        and not IS_MIRO_INIT
                        and MIRO_GDX_IN
                    ):
                        miro_load = miro.get_load_input_str(
                            loadable.name, gdx_in
                        )
                        strings.append(miro_load)
                    else:
                        strings.append(f"$loadDC {loadable.name}")

                strings.append("$gdxIn")

        strings += ["$offUNDF", "$offMulti"]

        if not IS_MIRO_INIT and MIRO_GDX_OUT:
            strings.append(miro.get_unload_output_str(self))

        gams_string = "\n".join(strings)

        if self._debugging_level == "keep":
            self._gams_string += gams_string + "\n"

        return gams_string

    def _load_records_from_gdx(
        self, load_from: str, names: Iterable[str]
    ) -> None:
        self._temp_container.read(load_from, names)
        original_state = self._options.miro_protect
        self._options.miro_protect = False

        for name in names:
            if name in self.data:
                updated_records = self._temp_container[name].records
                self[name].records = updated_records
                self[name].domain_labels = self[name].domain_names
            else:
                self._read(load_from, [name])

        self._options.miro_protect = original_state
        self._temp_container.data = {}

    def _load_records_with_rename(
        self, load_from: str, names: dict[str, str]
    ) -> None:
        self._temp_container.read(load_from, list(names.keys()))
        original_state = self._options.miro_protect
        self._options.miro_protect = False

        for gdx_name, gamspy_name in names.items():
            if gamspy_name in self.data:
                updated_records = self._temp_container[gdx_name].records
                self[gamspy_name].records = updated_records
                self[gamspy_name].domain_labels = self[
                    gamspy_name
                ].domain_names
            else:
                raise ValidationError(
                    f"Invalid renaming. `{gamspy_name}` does not exist in the container."
                )

        self._options.miro_protect = original_state
        self._temp_container.data = {}

    def _read(
        self,
        load_from: str | Container | gt.Container,
        symbol_names: list[str] | None = None,
        load_records: bool = True,
        mode: str | None = None,
        encoding: str | None = None,
    ) -> None:
        super().read(load_from, symbol_names, load_records, mode, encoding)
        self._cast_symbols(symbol_names)

    def read(
        self,
        load_from: str | Container | gt.Container,
        symbol_names: list[str] | None = None,
        load_records: bool = True,
        mode: str | None = None,
        encoding: str | None = None,
    ) -> None:
        """
        Reads specified symbols from the GDX file. If symbol_names are
        not provided, it reads all symbols from the GDX file.

        Parameters
        ----------
        load_from : str, Container, gt.Container
        symbol_names : List[str], optional
        load_records : bool
        mode : str, optional
        encoding : str, optional

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1', 'i2'])
        >>> m.write("test.gdx")
        >>> new_container = gp.Container()
        >>> new_container.read("test.gdx")
        >>> new_container.data.keys() == m.data.keys()
        True

        """
        self._read(load_from, symbol_names, load_records, mode, encoding)
        self._synch_with_gams()

    def write(
        self,
        write_to: str,
        symbol_names: list[str] | None = None,
        compress: bool = False,
        mode: str | None = None,
        eps_to_zero: bool = True,
    ) -> None:
        """
        Writes specified symbols to the GDX file. If symbol_names are
        not provided, it writes all symbols to the GDX file.

        Parameters
        ----------
        write_to : str
        symbol_names : List[str], optional
        compress : bool
        mode : str, optional
        eps_to_zero : bool

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1', 'i2'])
        >>> m.write("test.gdx")

        """
        super().write(
            write_to,
            symbol_names,
            compress,
            mode=mode,
            eps_to_zero=eps_to_zero,
        )

    def generateGamsString(self, show_raw: bool = False) -> str:
        """
        Generates the GAMS code

        Parameters
        ----------
        show_raw : bool, optional
            Shows the raw model without data and other necessary
            GAMS statements, by default False.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container(debugging_level="keep")
        >>> i = gp.Set(m, name="i")
        >>> gams_code = m.generateGamsString()

        """
        if self._debugging_level != "keep":
            raise ValidationError(
                "`debug_level` argument of the container must be set to 'keep' to use this function."
            )

        if not show_raw:
            return self._gams_string

        return utils._filter_gams_string(self._gams_string)

    def loadRecordsFromGdx(
        self,
        load_from: str,
        symbol_names: Iterable[str] | dict[str, str] | None = None,
    ) -> None:
        """
        Loads data of the given symbols from a GDX file. If no
        symbol names are given, data of all symbols are loaded.

        Parameters
        ----------
        load_from : str
            Path to the GDX file
        symbol_names : Iterable[str], dict[str, str], optional
            Symbol names whose data will be load from GDX, by default None.
            Default option loads records of all symbols in the GDX file.
            If given as a dict, keys are the symbol names in the GDX file, and
            values are the names of the GAMSPy symbols.

        Examples
        --------
        >>> from gamspy import Container, Set
        >>> m = Container()
        >>> i = Set(m, "i", records=["i1", "i2"])
        >>> m.write("test.gdx")
        >>> m2 = Container()
        >>> m2.loadRecordsFromGdx("test.gdx")
        >>> print(i.records.equals(m2["i"].records))
        True

        """
        if symbol_names is None:
            # If no symbol names are given, all records in the gdx should be loaded
            symbol_names = utils._get_symbol_names_from_gdx(
                self.system_directory, load_from
            )

        if isinstance(symbol_names, dict):
            self._load_records_with_rename(load_from, symbol_names)
        else:
            self._load_records_from_gdx(load_from, symbol_names)

        self._synch_with_gams()

    def addGamsCode(self, gams_code: str) -> None:
        """
        Adds an arbitrary GAMS code to the generate .gms file.
        Using addGAMSCode might result in a license error if no GAMSpy++ license is used.

        Parameters
        ----------
        gams_code : str
            GAMS code that you want to insert.

        Examples
        --------
        >>> from gamspy import Container
        >>> m = Container()
        >>> m.addGamsCode("scalar piHalf / [pi/2] /;")
        >>> m["piHalf"].toValue()
        np.float64(1.5707963267948966)

        """
        self._add_statement(gams_code)
        self._synch_with_gams(relaxed_domain_mapping=True, gams_to_gamspy=True)

        for _, symbol in self:
            symbol.modified = False
        self._unsaved_statements = []

    def close(self) -> None:
        """
        Stops the socket and releases resources. The container should not be used afterwards
        to communicate with the GAMS execution engine, e.g. creating new symbols, changing data,
        solves, etc. The container data (Container.data) is still available for read operations.
        """
        self._release_resources(
            self._is_socket_open, self._socket, self._process
        )
        self._is_socket_open = False

    def addAlias(
        self,
        name: str | None = None,
        alias_with: Set | Alias = None,  # type: ignore
    ) -> Alias:
        """
        Creates a new Alias and adds it to the container

        Parameters
        ----------
        name : str, optional
            Name of the alias.
        alias_with : Set | Alias
            Alias set object.

        Returns
        -------
        Alias

        Raises
        ------
        TypeError
            In case the alias_with is different than a Set or an Alias
        ValueError
            If there is symbol with same name but different type in the
            Container

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = m.addSet("i")
        >>> a = m.addAlias("a", i)

        """
        return gp.Alias(self, name, alias_with)

    def addSet(
        self,
        name: str | None = None,
        domain: Sequence[Set | Alias | str] | Set | Alias | str | None = None,
        is_singleton: bool = False,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_input: bool = False,
        is_miro_output: bool = False,
    ) -> Set:
        """
        Creates a Set and adds it to the container

        Parameters
        ----------
        name : str, optional
            Name of the set. Name is autogenerated by default.
        domain : list[Set | Alias | str] | Set | Alias | str, optional
            Domain of the set.
        is_singleton : bool, optional
            Whether the set is a singleton set. Singleton sets cannot contain more than one element.
        records : pd.DataFrame | np.ndarray | list, optional
            Records of the set.
        domain_forwarding : bool, optional
            Whether the set forwards the domain.
        description : str, optional
            Description of the set.
        uels_on_axes : bool
            Assume that symbol domain information is contained in the axes of the given records.
        is_miro_input : bool
            Whether the symbol is a GAMS MIRO input symbol. See: https://gams.com/miro/tutorial.html
        is_miro_output : bool
            Whether the symbol is a GAMS MIRO output symbol. See: https://gams.com/miro/tutorial.html

        Returns
        -------
        Set

        Raises
        ------
        err
            In case arguments are not valid
        ValueError
            When there is symbol with same name in the
            Container

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = m.addSet("i")

        """
        return gp.Set(
            self,
            name,
            domain,
            is_singleton,
            records,
            domain_forwarding,
            description,
            uels_on_axes,
            is_miro_input=is_miro_input,
            is_miro_output=is_miro_output,
        )

    def addParameter(
        self,
        name: str | None = None,
        domain: Sequence[Set | Alias | str] | Set | Alias | str | None = None,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_input: bool = False,
        is_miro_output: bool = False,
        is_miro_table: bool = False,
    ) -> Parameter:
        """
        Creates a Parameter and adds it to the Container

        Parameters
        ----------
        name : str, optional
            Name of the parameter. Name is autogenerated by default.
        domain : list[Set | Alias | str] | Set | Alias | str, optional
            Domain of the parameter.
        records : int | float | pd.DataFrame | np.ndarray | list, optional
            Records of the parameter.
        domain_forwarding : bool, optional
            Whether the parameter forwards the domain.
        description : str, optional
            Description of the parameter.
        uels_on_axes : bool
            Assume that symbol domain information is contained in the axes of the given records.
        is_miro_input : bool
            Whether the symbol is a GAMS MIRO input symbol. See: https://gams.com/miro/tutorial.html
        is_miro_output : bool
            Whether the symbol is a GAMS MIRO output symbol. See: https://gams.com/miro/tutorial.html
        is_miro_table : bool
            Whether the symbol is a GAMS MIRO table symbol. See: https://gams.com/miro/tutorial.html

        Returns
        -------
        Parameter

        Raises
        ------
        err
            In case arguments are not valid
        ValueError
            If there is symbol with same name but different type in the
            Container

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> a = m.addParameter("a")

        """
        return gp.Parameter(
            self,
            name,
            domain,
            records,
            domain_forwarding,
            description,
            uels_on_axes,
            is_miro_input=is_miro_input,
            is_miro_output=is_miro_output,
            is_miro_table=is_miro_table,
        )

    def addVariable(
        self,
        name: str | None = None,
        type: str = "free",
        domain: Sequence[Set | Alias | str] | Set | Alias | str | None = None,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_output: bool = False,
    ) -> Variable:
        """
        Creates a Variable and adds it to the Container

        Parameters
        ----------
        name : str, optional
            Name of the variable. Name is autogenerated by default.
        type : str, optional
            Type of the variable. "free" by default.
        domain : list[Set | Alias | str] | Set | Alias | str, optional
            Domain of the variable.
        records : Any, optional
            Records of the variable.
        domain_forwarding : bool, optional
            Whether the variable forwards the domain.
        description : str, optional
            Description of the variable.
        is_miro_output : bool
            Whether the symbol is a GAMS MIRO output symbol. See: https://gams.com/miro/tutorial.html

        Returns
        -------
        Variable

        Raises
        ------
        err
            In case arguments are not valid
        ValueError
            If there is symbol with same name but different type in the
            Container

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> v = m.addVariable("v")

        """
        return gp.Variable(
            self,
            name,
            type,
            domain,
            records,
            domain_forwarding,
            description,
            uels_on_axes,
            is_miro_output=is_miro_output,
        )

    def addEquation(
        self,
        name: str | None = None,
        type: str | EquationType = "regular",
        domain: Sequence[Set | Alias | str] | Set | Alias | str | None = None,
        definition: Variable | Operation | Expression | None = None,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_output: bool = False,
        definition_domain: list | None = None,
    ) -> Equation:
        """
        Creates an Equation and adds it to the Container

        Parameters
        ----------
        name : str, optional
            Name of the equation. Name is autogenerated by default.
        type : str
            Type of the equation. "regular" by default.
        domain : list[Set | Alias | str] | Set | Alias | str, optional
            Domain of the variable.
        definition: Expression, optional
            Definition of the equation.
        records : Any, optional
            Records of the equation.
        domain_forwarding : bool, optional
            Whether the equation forwards the domain.
        description : str, optional
            Description of the equation.
        uels_on_axes: bool
            Assume that symbol domain information is contained in the axes of the given records.
        definition_domain: list, optional
            Definiton domain of the equation.
        is_miro_output : bool
            Whether the symbol is a GAMS MIRO output symbol. See: https://gams.com/miro/tutorial.html

        Returns
        -------
        Equation

        Raises
        ------
        err
            In case arguments are not valid
        ValueError
            If there is symbol with same name but different type in the
            Container

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = m.addEquation("i")

        """
        return gp.Equation(
            self,
            name,
            type,
            domain,
            definition,
            records,
            domain_forwarding,
            description,
            uels_on_axes,
            is_miro_output,
            definition_domain,
        )

    def addModel(
        self,
        name: str | None = None,
        description: str = "",
        problem: Problem | str = Problem.MIP,
        equations: Sequence[Equation] = [],
        sense: Sense | str = Sense.FEASIBILITY,
        objective: Variable | Expression | None = None,
        matches: dict[
            Equation | Sequence[Equation],
            Variable | Sequence[Variable],
        ]
        | None = None,
        limited_variables: Sequence[Variable] | None = None,
        external_module: str | None = None,
    ) -> Model:
        """
        Creates a Model and adds it to the Container

        Parameters
        ----------
        name : str, optional
            Name of the model. Name is autogenerated by default.
        description : str, optional
            Description of the model.
        equations : Sequence[Equation]
            Sequence of Equation objects.
        problem : Problem or str, optional
            'LP', 'NLP', 'QCP', 'DNLP', 'MIP', 'RMIP', 'MINLP', 'RMINLP', 'MIQCP', 'RMIQCP', 'MCP', 'CNS', 'MPEC', 'RMPEC', 'EMP', or 'MPSGE',
            by default Problem.LP.
        sense : Sense, optional
            "MIN", "MAX", or "FEASIBILITY".
        objective : Variable | Expression, optional
            Objective variable to minimize or maximize or objective itself.
        matches : dict[Equation | Sequence[Equation], Variable | Sequence[Variable]], optional
            Equation - Variable matches for MCP models.
        limited_variables : Sequence, optional
            Allows limiting the domain of variables used in a model.
        external_module: str, optional
            The name of the external module in which the external equations are implemented

        Returns
        -------
        Model

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> e = gp.Equation(m, "e")
        >>> model = m.addModel("my_model", "LP", [e])

        """
        return gp.Model(
            self,
            name,
            description,
            problem,
            equations,
            sense,
            objective,
            matches,
            limited_variables,
            external_module=external_module,
        )

    def copy(self, working_directory: str) -> Container:
        """
        Creates a copy of the Container. Should not be invoked after
        creating the model.

        Parameters
        ----------
        working_directory : str, optional
            Working directory of the new Container, by default None

        Returns
        -------
        Container

        Raises
        ------
        ValidationError

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i")
        >>> new_cont = m.copy(working_directory="test")
        >>> new_cont.data.keys() == m.data.keys()
        True

        """
        os.makedirs(working_directory, exist_ok=True)
        m = Container(working_directory=working_directory)
        if m.working_directory == self.working_directory:
            raise ValidationError(
                "Copy of a container cannot have the same working directory"
                " with the original container."
            )

        self.write(m._job + "in.gdx")
        m._read(m._job + "in.gdx")

        # if already defined equations exist, add them to .gms file
        for equation in self.getEquations():
            if equation._definition is not None:
                m._add_statement(equation._definition)
                m[equation.name]._definition = equation._definition
                m._synch_with_gams()

        return m

    def serialize(self, path: str) -> None:
        """
        Serializes the Container into a zip file.

        Parameters
        ----------
        path : str
            Path to the zip file.
        """
        gp.serialize(self, path)

    def getEquations(self) -> list[Equation]:
        """
        Returns all equation symbols in the Container.

        Returns
        -------
        list[Equation]

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> eq1 = gp.Equation(m, name="eq1")
        >>> eq2 = gp.Equation(m, name="eq2")
        >>> equation_objects = m.getEquations()

        """
        equations = [
            equation
            for equation in self.listEquations()
            if not equation.startswith(gp.Model._generate_prefix)
        ]
        return self.getSymbols(equations)

    def importExtrinsicLibrary(
        self, lib_path: str, functions: dict[str, str]
    ) -> ExtrinsicLibrary:
        """
        Imports an extrinsic library to the GAMS environment.

        Parameters
        ----------
        lib_path : str
            Path to the .so, .dylib or .dll file that contains the extrinsic library
        functions : dict[str, str]
            Names of the functions as a dictionary. Key is the desired function name in GAMSPy
            and value is the function name in the extrinsic library.

        Returns
        -------
        ExtrinsicLibrary

        Raises
        ------
        FileNotFoundError
            In case the extrinsic library does not exist in the given path.
        """
        if not os.path.exists(lib_path):
            raise FileNotFoundError(f"`{lib_path}` is not a valid path.")

        external_lib = ExtrinsicLibrary(lib_path, functions)
        self._add_statement(external_lib)

        return external_lib

    def gamsJobName(self) -> str | None:
        """
        Returns the name of the latest GAMS job that was executed

        Returns
        -------
        str | None

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=["seattle", "san-diego"], description="canning plants")
        >>> gams_file_name = f"{m.gamsJobName()}.gms"

        """
        return self._job

    def gdxInputPath(self) -> str:
        """
        Path to the input GDX file

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=["seattle", "san-diego"], description="canning plants")
        >>> gdx_path = m.gdxInputPath()

        """
        return self._gdx_in

    def gdxOutputPath(self) -> str:
        """
        Path to the output GDX file

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego"], description="canning plants")
        >>> ii = gp.Set(m, name="ii", domain=i, description="seattle plant")
        >>> ii['seattle'] = True
        >>> gdx_path = m.gdxOutputPath()

        """
        return self._gdx_out



from __future__ import annotations

import base64
import copy
import io
import json
import logging
import os
import shutil
import tempfile
import time
import urllib.parse
import zipfile
from typing import TYPE_CHECKING

import certifi
import urllib3

import gamspy._backend.backend as backend
import gamspy.utils as utils
from gamspy._options import Options
from gamspy.exceptions import (
    EngineClientException,
    EngineException,
    GamspyException,
    ValidationError,
)

if TYPE_CHECKING:
    from gamspy import Container, Model
    from gamspy._symbols.symbol import Symbol


logger = logging.getLogger("ENGINE")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


MAX_REQUEST_ATTEMPS = 3
ZIP_NAME = "data.zip"
INEX_FILE_NAME = "inex.json"
STATUS_MAP = {
    -10: "waiting",
    -3: "cancelled",
    -2: "cancelling",
    -1: "corrupted",
    0: "queued",
    1: "running",
    2: "outputting",
    10: "finished",
}


def get_relative_paths(paths: list[str], start: str) -> list[str]:
    relative_paths = []
    for path in paths:
        relative_path = os.path.relpath(path, start)
        if relative_path.startswith(f"..{os.sep}"):
            raise ValidationError(
                "Extra model file path must be relative to the working"
                f" directory. The given path: {path}, the working"
                f" directory: {start}, the"
                f" relative path: {relative_path}"
            )

        relative_paths.append(relative_path)

    return relative_paths


SCOPES = [
    "READONLY",
    "NAMESPACES",
    "JOBS",
    "USERS",
    "HYPERCUBE",
    "CLEANUP",
    "LICENSES",
    "USAGE",
    "AUTH",
    "CONFIGURATION",
]


class EngineConfiguration:
    """Configuration that allows the execution of jobs on a specific GAMS Engine instance"""

    def __init__(
        self, host: str, username: str, password: str, jwt: str, namespace: str
    ):
        self.jwt = jwt
        self.username = username
        self.password = password
        self.namespace = namespace
        self.host = self._verify_host(host)

    def _verify_host(self, host: str):
        validated_url = urllib.parse.urlparse(host)
        if validated_url.scheme not in ("http", "https"):
            raise ValidationError(
                "Invalid GAMS Engine host. Only HTTP and HTTPS protocols supported"
            )
        if validated_url.netloc == "":
            raise ValidationError(
                "Invalid GAMS Engine host. Make sure you provide a valid URL."
            )
        host = host.rstrip("/")
        if not host.endswith("/api"):
            host += "/api"

        return host

    def _get_auth_header(self):
        """Returns authentication header"""
        if not self.username:
            if not self.jwt:
                raise ValidationError(
                    "Neither username/password nor JWT token provided for authentication with the GAMS Engine instance."
                )
            return "Bearer " + self.jwt

        return "Basic " + base64.b64encode(
            (self.username + ":" + self.password).encode("utf-8")
        ).decode("utf-8")


class Endpoint: ...


class Auth(Endpoint):
    def __init__(self, client: EngineClient) -> None:
        self.client = client
        self._http = client._http
        self.engine_options = client.engine_options

    def post(
        self,
        expires_in: int = 14400,
        scope: list[str] | None = None,
    ) -> str:  # type: ignore
        """
        Creates a JSON Web Token(JWT) for authentication

        Parameters
        ----------
        expires_in : int, optional
            Expiration time, by default 14400
        scope : list[str] | None, optional
            Scope of the token, by default None

        Returns
        -------
        str
            token

        Raises
        ------
        EngineClientException
            In case bad request
        EngineClientException
            In case unauthorized request
        EngineClientException
            In case there is an internal error
        GamspyException
            In case the status code is unrecognized
        """
        info = {"expires_in": str(expires_in)}

        if isinstance(scope, list):
            for elem in scope:
                if elem not in SCOPES:
                    raise ValidationError(f"{elem} is not a valid scope")

            scope_info = " ".join(scope)
            info.update({"scope": scope_info})

        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "POST",
                self.client._engine_config.host
                + "/auth/?"
                + urllib.parse.urlencode(info, doseq=True),
                headers=self.client.get_request_headers(),
            )

            response_data = r.data.decode("utf-8", errors="replace")

            if r.status in (200, 400, 401, 500):
                info = json.loads(response_data)

            if r.status == 200:
                return info["token"]
            elif r.status == 400:
                raise EngineClientException(f"Bad request: {info['message']}")
            elif r.status == 401:
                raise EngineClientException("Unauthorized!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue
            elif r.status == 500:
                raise EngineClientException(
                    f"Internal error: {info['message']}"
                )

        raise GamspyException(f"Unrecognized status code {r.status}")

    def login(
        self, expires_in: int = 14400, scope: list[str] | None = None
    ) -> str:
        """
        Creates a JSON Web Token(JWT) for authentication (username and password in request body)

        Parameters
        ----------
        expires_in : int, optional
            Expiration time for the token, by default 14400
        scope: list[str], optional
            Scope of the token, by default None

        Returns
        -------
        str
        """
        info = {
            "username": self.client._engine_config.username,
            "password": self.client._engine_config.password,
            "expires_in": expires_in,
        }

        if isinstance(scope, list):
            for elem in scope:
                if elem not in SCOPES:
                    raise ValidationError(f"{elem} is not a valid scope")

            scope_info = " ".join(scope)
            info.update({"scope": scope_info})

        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "POST",
                self.client._engine_config.host + "/auth/login",
                fields=info,
            )

            response_data = r.data.decode("utf-8", errors="replace")
            if r.status in (200, 400, 401, 500):
                info = json.loads(response_data)

            if r.status == 200:
                return info["token"]  # type: ignore
            elif r.status == 400:
                raise EngineClientException(f"Bad request: {info['message']}")
            elif r.status == 401:
                raise EngineClientException(f"Unauthorized: {info['message']}")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue
            elif r.status == 500:
                raise EngineClientException(
                    f"Internal error: {info['message']}"
                )

        raise GamspyException(f"Unrecognized status code {r.status}")

    def logout(self) -> str:
        """
        Invalidates all of your JSON Web Tokens(JWTs)

        Returns
        -------
        str
        """
        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "POST",
                self.client._engine_config.host + "/auth/logout",
                headers=self.client.get_request_headers(),
            )

            if r.status == 200:
                response_data = r.data.decode("utf-8", errors="replace")
                info = json.loads(response_data)
                return info["message"]
            elif r.status == 400:
                raise EngineClientException("Bad request!")
            elif r.status == 401:
                raise EngineClientException("Unauthorized!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue
            elif r.status == 500:
                raise EngineClientException("Internal error!")

        raise GamspyException(f"Unrecognized status code {r.status}")


class Job(Endpoint):
    def __init__(self, client: EngineClient) -> None:
        self.client = client
        self._http = client._http
        self.engine_options = client.engine_options

    def get(self, token: str) -> tuple[int, str, int | None]:
        """
        Get request to /jobs/{token} which returns the details of a job.
        Refer to https://engine.gams.com/api/ for more details.

        Parameters
        ----------
        token : str
            Job token

        Returns
        -------
        tuple[int, str, int]
            Job status, job status message, and gams exit code

        Raises
        ------
        EngineClientException
            If get request has failed.
        """
        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "GET",
                self.client._engine_config.host + f"/jobs/{token}",
                headers=self.client.get_request_headers(),
            )
            response_data = r.data.decode("utf-8", errors="replace")

            if r.status == 200:
                info = json.loads(response_data)
                job_status = int(info["status"])
                return (
                    job_status,
                    STATUS_MAP[job_status],
                    info["process_status"],
                )
            elif r.status == 403:
                raise EngineClientException("Unauthorized!")
            elif r.status == 404:
                raise EngineClientException(f"Job {token} not found!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue
            elif r.status == 500:
                raise EngineClientException("Internal error!")

        raise EngineClientException(
            "Creating job on GAMS Engine failed after: "
            + str(MAX_REQUEST_ATTEMPS)
            + " attempts. Message: "
            + response_data
        )

    def post(
        self,
        working_directory: str,
        gms_file: str,
        pf_file: str | None = None,
    ) -> str:
        """
        Post request to /jobs which submits a new job to be solved.
        Refer to https://engine.gams.com/api/ for more details.

        Parameters
        ----------
        working_directory : str
            Working directory
        gms_file : str
            Name of the gms file
        pf_file: str | None
            Name of the pf file

        Returns
        -------
        str
            Token

        Raises
        ------
        EngineClientException
            If post request has failed.
        """
        model_data_zip = self._create_zip_file(
            working_directory, gms_file, pf_file
        )
        gms_file = os.path.relpath(gms_file, working_directory)
        pf_file = (
            os.path.relpath(pf_file, working_directory)
            if pf_file is not None
            else None
        )
        query_params, file_params = self._get_params(
            model_data_zip, gms_file, pf_file
        )

        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "POST",
                self.client._engine_config.host
                + "/jobs/?"
                + urllib.parse.urlencode(query_params, doseq=True),
                fields=file_params,
                headers=self.client.get_request_headers(),
            )
            response_data = r.data.decode("utf-8", errors="replace")
            if r.status == 201:
                return json.loads(response_data)["token"]
            elif r.status == 400:
                raise EngineClientException("Input is not valid!")
            elif r.status == 401:
                raise EngineClientException("Invalid authentication!")
            elif r.status == 402:
                raise EngineClientException("Quota exceeded!")
            elif r.status == 403:
                raise EngineClientException("Unauthorized access!")
            elif r.status == 404:
                raise EngineClientException("Namespace could not be found!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue
            elif r.status == 500:
                raise EngineClientException("Internal error!")

        raise EngineClientException(
            "Creating job on GAMS Engine failed after: "
            + str(MAX_REQUEST_ATTEMPS)
            + " attempts. Message: "
            + response_data
        )

    def get_results(self, token: str, working_directory: str):
        """
        Get request to /jobs/{token}/result which downloads the job results.
        Downloaded results are unpacked to working directory.
        Refer to https://engine.gams.com/api/ for more details.

        Parameters
        ----------
        token : str
            Job token
        working_directory : str
            Working directory

        Raises
        ------
        EngineClientException
            If get request has failed.
        """
        if not os.path.exists(working_directory):
            os.makedirs(working_directory, exist_ok=True)

        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "GET",
                self.client._engine_config.host + f"/jobs/{token}/result",
                headers=self.client.get_request_headers(),
                preload_content=False,
            )

            if r.status == 200:
                fd, path = tempfile.mkstemp()

                try:
                    with open(path, "wb") as out:
                        while True:
                            data = r.read(6000)
                            if not data:
                                break
                            out.write(data)
                            out.flush()

                    r.release_conn()

                    with zipfile.ZipFile(path, "r") as zip_ref:
                        zip_ref.extractall(working_directory)
                finally:
                    os.close(fd)
                    os.remove(path)

                logger.info(
                    f"Results have been extracted to your working directory: {working_directory}."
                )
            elif r.status == 400:
                raise EngineClientException("Bad request!")
            elif r.status == 401:
                raise EngineClientException("Unauthorized!")
            elif r.status == 403:
                raise EngineClientException("Job data does not exist!")
            elif r.status == 404:
                raise EngineClientException("No job found!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue
            else:
                response_data = r.data.decode("utf-8", errors="replace")
                raise EngineClientException(
                    "Fatal error while getting the results back from engine. GAMS"
                    f" Engine return code: {r.status}. Error message:"
                    f" {response_data}"
                )

    def delete_results(self, token: str):
        """
        Delete request to /jobs/{token} which deletes the job results.
        Refer to https://engine.gams.com/api/ for more details.

        Parameters
        ----------
        token : str
            Job token

        Raises
        ------
        EngineClientException
            If job data does not exist in GAMS Engine.
        EngineClientException
            If delete request has failed.
        """
        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "DELETE",
                self.client._engine_config.host + "/jobs/" + token + "/result",
                headers=self.client.get_request_headers(),
            )
            response_data = r.data.decode("utf-8", errors="replace")

            if r.status == 200:
                logger.info(f"Results for {token} have been deleted.")
                return
            elif r.status == 400:
                raise EngineClientException("Bad request!")
            elif r.status == 401:
                raise EngineClientException("Unauthorized!")
            elif r.status == 403:
                raise EngineClientException("Job data does not exist!")
            elif r.status == 404:
                raise EngineClientException("No job found!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue

        raise EngineClientException(
            "Removing job result failed with status code: "
            + str(r.status)
            + ". Message: "
            + response_data
        )

    def get_logs(self, token: str) -> tuple[str, bool]:
        """
        Get request to /jobs/{token}/unread-logs which returns stdout of a job.
        Refer to https://engine.gams.com/api/ for more details.

        Parameters
        ----------
        token : str
            Job token

        Returns
        -------
        tuple[str, bool]
            Current output buffer and queue finished status

        Raises
        ------
        EngineClientException
            If get request has failed.
        """
        for attempt_number in range(MAX_REQUEST_ATTEMPS):
            r = self._http.request(
                "DELETE",
                self.client._engine_config.host + f"/jobs/{token}/unread-logs",
                headers=self.client.get_request_headers(),
            )
            response_data = r.data.decode("utf-8", errors="replace")

            if r.status == 200:
                info = json.loads(response_data)
                return info["message"], info["queue_finished"]
            elif r.status == 308:
                info = json.loads(response_data)
                return info["message"], True
            elif r.status == 403:
                raise EngineClientException(
                    "Cannot get logs of pending job / Unauthorized access!"
                )
            elif r.status == 404:
                raise EngineClientException("No job found!")
            elif r.status == 429:
                time.sleep(2**attempt_number)  # retry with exponential backoff
                continue

        raise EngineClientException(
            "Getting logs failed with status code: "
            + str(r.status)
            + ". "
            + response_data
            + "."
        )

    def _create_zip_file(
        self,
        working_directory: str,
        gms_file: str,
        pf_file: str | None,
    ) -> io.BytesIO:
        model_data_zip = io.BytesIO()
        model_files = [gms_file]
        if pf_file is not None:
            model_files.append(pf_file)

        model_files += self.client.extra_model_files
        model_files = get_relative_paths(model_files, working_directory)

        with zipfile.ZipFile(
            model_data_zip, "w", zipfile.ZIP_DEFLATED
        ) as model_data:
            for model_file in model_files:
                model_data.write(
                    os.path.join(working_directory, model_file),
                    arcname=model_file,
                )

        model_data_zip.seek(0)

        return model_data_zip

    def _get_params(self, model_data_zip, gms_file, pf_file: str | None):
        file_params = {}
        query_params = (
            copy.deepcopy(self.engine_options) if self.engine_options else {}
        )

        query_params["namespace"] = self.client._engine_config.namespace

        if "data" in query_params or "model_data" in query_params:
            raise ValidationError(
                "`engine_options` must not include keys `data` or "
                "`model_data`. Please use `extra_model_files` to "
                "provide additional files to send to GAMS Engine.",
            )

        if "inex_file" in query_params:
            if isinstance(query_params["inex_file"], io.IOBase):
                file_params["inex_file"] = (
                    INEX_FILE_NAME,
                    query_params["inex_file"].read(),
                    "application/json",
                )
            else:
                with open(query_params["inex_file"], "rb") as f:
                    file_params["inex_file"] = (
                        INEX_FILE_NAME,
                        f.read(),
                        "application/json",
                    )
            del query_params["inex_file"]

        if "model" in query_params:
            file_params["data"] = (
                ZIP_NAME,
                model_data_zip.read(),
                "application/zip",
            )
        else:
            query_params["run"] = gms_file
            query_params["model"] = os.path.splitext(gms_file)[0]
            file_params["model_data"] = (
                ZIP_NAME,
                model_data_zip.read(),
                "application/zip",
            )

        model_data_zip.close()

        if "arguments" in query_params:
            if not isinstance(query_params["arguments"], list):
                query_params["arguments"] = [query_params["arguments"]]

            if pf_file is not None:
                query_params["arguments"].append(f"pf={pf_file}")
        else:
            if pf_file is not None:
                query_params["arguments"] = [f"pf={pf_file}"]

        return query_params, file_params


class EngineClient:
    """
    GAMS Engine Client.

    Parameters
    ----------
    host : str
        Host address.
    username : str, optional
        Username
    password : str, optional
        Password
    jwt : str, optional
        JWT Token
    namespace : str, optional
        Namespace, by default "global"
    extra_model_files : list[str], optional
        extra_model_files, by default []
    engine_options : dict | None, optional
        GAMS Engine options, by default None
    remove_results : bool, optional
        Whether to remove the results from GAMS Engine Server, by default False
    is_blocking : bool, optional
        Synchronous (blocking) or asynchronous, by default True

    Examples
    --------
    >>> import gamspy as gp
    >>> client = gp.EngineClient(host="https://engine.gams.com/api/", username="<username>", password="<password>")

    """

    def __init__(
        self,
        host: str,
        username: str = "",
        password: str = "",
        jwt: str = "",
        namespace: str = "global",
        extra_model_files: list[str] = [],
        engine_options: dict | None = None,
        remove_results: bool = False,
        is_blocking: bool = True,
    ):
        self.host = host
        self.username = username
        self.password = password
        self.jwt = jwt
        self.namespace = namespace
        self.extra_model_files = extra_model_files
        self.engine_options = engine_options
        self.remove_results = remove_results
        self.is_blocking = is_blocking
        self.tokens: list[str] = []

        self._http = urllib3.PoolManager(
            cert_reqs="CERT_REQUIRED", ca_certs=certifi.where()
        )

        self._engine_config = EngineConfiguration(
            self.host,
            self.username,
            self.password,
            self.jwt,
            self.namespace,
        )

        # Endpoints
        self.job = Job(self)
        self.auth = Auth(self)

    def get_request_headers(self):
        return {
            "Authorization": self._engine_config._get_auth_header(),
            "User-Agent": "GAMSPy EngineClient",
            "Accept": "application/json",
        }


class GAMSEngine(backend.Backend):
    def __init__(
        self,
        container: Container,
        client: EngineClient | None,
        options: Options,
        solver: str,
        solver_options: dict | None,
        output: io.TextIOWrapper | None,
        model: Model,
        load_symbols: list[Symbol] | None,
    ) -> None:
        if client is None or not isinstance(client, EngineClient):
            raise ValidationError(
                "`engine_client` must be provided to solve on GAMS Engine"
            )

        if solver.lower() == "mpsge":
            raise ValidationError(
                f"`{solver}` is not a valid solver for GAMS Engine."
            )

        super().__init__(
            container,
            model,
            options,
            solver,
            solver_options,
            output,
            load_symbols,
        )
        self.client = client

        self.job_name = self.get_job_name()
        self.gms_file = self.job_name + ".gms"
        self.pf_file = self.job_name + ".pf"
        self.restart_file = self.job_name + ".g00"
        self.trace_file = self.job_name + ".txt"

    def is_async(self):
        return not self.client.is_blocking

    def run(
        self,
        relaxed_domain_mapping: bool = False,
        gams_to_gamspy: bool = False,
    ):
        # Run a dummy job to get the restart file to be sent to GAMS Engine
        self._create_restart_file()

        # Generate gams string and write modified symbols to gdx
        gams_string = self.preprocess(os.path.basename(self.container._gdx_in))

        self.execute_gams(gams_string)

        if self.is_async():
            return None

        # Synchronize GAMSPy with checkpoint and return a summary
        summary = self.postprocess()

        # Run another dummy job to synchronize GAMS and GAMSPy state
        self._sync()

        return summary

    def execute_gams(self, gams_string: str):
        extra_options = {
            "gdx": os.path.basename(self.container._gdx_out),
            "gdxSymbols": "newOrChanged",
            "trace": os.path.basename(self.trace_file),
            "restart": os.path.basename(self.restart_file),
            "input": os.path.basename(self.gms_file),
        }
        self.options._set_extra_options(extra_options)
        self.options._export(self.pf_file, self.output)

        with open(self.gms_file, "w", encoding="utf-8") as file:
            file.write(gams_string)

        try:
            original_extra_files = copy.deepcopy(self.client.extra_model_files)
            self.client.extra_model_files = self._append_gamspy_files()
            token = self.client.job.post(
                self.container.working_directory,
                self.gms_file,
                self.pf_file,
            )
            self.client.extra_model_files = original_extra_files
            self.client.tokens.append(token)

            if not self.is_async():
                job_status, message, _ = self.client.job.get(token)

                if job_status not in STATUS_MAP:
                    raise EngineException(
                        "Unknown job status code! Currently supported job"
                        f" status codes: {STATUS_MAP.keys()}",
                        return_code=job_status,
                        status_code=job_status,
                    )

                if job_status in (-1, -3):
                    raise EngineException(
                        "Could not get job results because the job is"
                        f" {message}.",
                        return_code=job_status,
                        status_code=job_status,
                    )

                while job_status in (-10, -2, 0):
                    logger.info(f"Job status is {message}...")
                    job_status, _, _ = self.client.job.get(token)

                while job_status in (1, 2):
                    message, is_finished = self.client.job.get_logs(token)

                    if self.output is not None:
                        self.output.write(message)

                    if is_finished:
                        job_status = 10

                self.client.job.get_results(
                    token, self.container.working_directory
                )

                engine_output_path = os.path.join(
                    self.container.working_directory, "log_stdout.txt"
                )

                if self.options.log_file:
                    # logoption = 2 | 4
                    shutil.move(engine_output_path, self.options.log_file)
                else:
                    try:
                        os.remove(engine_output_path)  # logoption = 0 | 3
                    except FileNotFoundError:
                        ...

                assert self.model is not None
                self.model._update_model_attributes()

                if self.client.remove_results:
                    self.client.job.delete_results(token)
        finally:
            self.container._unsaved_statements = []
            self.container._delete_autogenerated_symbols()

    def postprocess(self):
        super().load_records()
        if self.client.remove_results or self.model is None:
            return None

        if not self.is_async():
            self.parse_listings()
            return self.prepare_summary(self.trace_file)

        return None

    def _prepare_dummy_options(self) -> dict:
        scrdir = self.container._process_directory

        extra_options = {
            "gdx": self.container._gdx_out,
            "gdxSymbols": "newOrChanged",
            "trace": self.trace_file,
            "input": self.gms_file,
            "sysdir": self.container.system_directory,
            "scrdir": scrdir,
            "scriptnext": os.path.join(scrdir, "gamsnext.sh"),
            "logoption": 0,
            "previouswork": 1,
            "license": utils._get_license_path(
                self.container.system_directory
            ),
        }

        if self.container._network_license:
            extra_options["netlicense"] = os.path.join(scrdir, "gamslice.dat")

        return extra_options

    def _create_restart_file(self):
        with open(self.gms_file, "w", encoding="utf-8") as gams_file:
            gams_file.write("")

        options = Options()
        extra_options = self._prepare_dummy_options()
        options._set_extra_options(extra_options)
        options._extra_options["save"] = self.restart_file
        options._export(self.pf_file)

        self.container._send_job(self.job_name, self.pf_file)

    def _sync(self):
        symbols = utils._get_symbol_names_from_gdx(
            self.container.system_directory, self.container._gdx_out
        )
        dirty_str = ",".join(symbols)
        with open(self.gms_file, "w", encoding="utf-8") as gams_file:
            gams_file.write(
                f'execute_load "{self.container._gdx_out}", {dirty_str};'
            )

        options = Options()
        extra_options = self._prepare_dummy_options()
        options._set_extra_options(extra_options)
        options._export(self.pf_file)

        self.container._send_job(self.job_name, self.pf_file)

    def _append_gamspy_files(self) -> list[str]:
        extra_model_files = self.client.extra_model_files + [self.restart_file]
        if os.path.exists(self.container._gdx_in):
            extra_model_files.append(self.container._gdx_in)

        if self.solver_options:
            assert self.solver is not None
            extra_model_files.append(
                os.path.join(
                    self.container.working_directory,
                    f"{self.solver.lower()}.opt",
                )
            )

        if self.options.loadpoint is not None:
            extra_model_files.append(self.options.loadpoint)

        assert self.model is not None
        if self.model.external_module is not None:
            extra_model_files.append(self.model.external_module)

        return extra_model_files



from __future__ import annotations

import json
import os
import sys
from typing import TYPE_CHECKING

import gamspy as gp
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Container, Equation, Parameter, Set, Variable

MIRO_GDX_IN = os.getenv("GAMS_IDC_GDX_INPUT", None)
MIRO_GDX_OUT = os.getenv("GAMS_IDC_GDX_OUTPUT", None)


def get_load_input_str(name: str, gdx_in: str) -> str:
    strings = [
        "$gdxIn",  # close the old one
        f"$gdxIn {MIRO_GDX_IN}",  # open the new one
        f"$loadDC {name}",
        "$gdxIn",  # close the new one
        f"$gdxIn {gdx_in}",  # reopen the previous one
    ]

    return "\n".join(strings)


def get_unload_output_str(container: Container) -> str:
    unload_str = ",".join(container._miro_output_symbols)
    return f"execute_unload '{MIRO_GDX_OUT}' {unload_str}"


def load_miro_symbol_records(container: Container):
    # Load records of miro input symbols
    if MIRO_GDX_IN and container._miro_input_symbols:
        names = [
            name
            for name in container._miro_input_symbols
            if not container[name]._already_loaded
        ]
        container._load_records_from_gdx(MIRO_GDX_IN, names)
        for name in names:
            symbol = container[name]
            symbol._already_loaded = True
            if isinstance(symbol, gp.Parameter) and symbol._is_miro_table:
                symbol._records.columns = symbol.domain_names + ["value"]

    # Load records of miro output symbols
    if MIRO_GDX_OUT and container._miro_output_symbols:
        container._load_records_from_gdx(
            MIRO_GDX_OUT, container._miro_output_symbols
        )


def _write_default_gdx_miro(container: Container) -> None:
    # create data_<model>/default.gdx
    symbols = container._miro_input_symbols + container._miro_output_symbols

    filename = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    directory = os.path.dirname(sys.argv[0])
    data_path = os.path.join(directory, f"data_{filename}")
    try:
        os.mkdir(data_path)
    except FileExistsError:
        pass

    container.write(os.path.join(data_path, "default.gdx"), symbols)


class MiroJSONEncoder:
    def __init__(
        self,
        container: Container,
    ):
        self.container = container
        self.model_title = "GAMSPy App"
        self.input_symbols = container._miro_input_symbols
        self.output_symbols = container._miro_output_symbols
        self.input_scalars = self._find_scalars(self.input_symbols)
        self.output_scalars = self._find_scalars(self.output_symbols)

    def _find_scalars(self, symbols: list[str]) -> list[str]:
        scalars = []
        for name in symbols:
            symbol = self.container[name]

            if len(symbol.domain) == 0:
                scalars.append(name)

            if isinstance(symbol, gp.Set) and symbol.is_singleton:
                scalars.append(name)

        return scalars

    def _prepare_scalars(
        self, alias: str, symbols: list[str]
    ) -> tuple[dict | None, dict | None]:
        sp_names, sp_texts, sp_types = [], [], []
        ve_names, ve_texts, ve_types = [], [], []

        for name in symbols:
            symbol = self.container[name]

            if isinstance(symbol, (gp.Set, gp.Parameter)):
                sp_names.append(name)
                sp_texts.append(
                    symbol.description if symbol.description else symbol.name
                )
                sp_types.append(type(symbol).__name__.lower())
            elif isinstance(symbol, (gp.Variable, gp.Equation)):
                ve_names.append(name)
                ve_texts.append(
                    symbol.description if symbol.description else symbol.name  # type: ignore
                )
                ve_types.append(type(symbol).__name__.lower())

        sp_scalars_dict = None
        ve_scalars_dict = None

        if len(sp_names) != 0:
            sp_scalars_dict = {
                "alias": alias,
                "symnames": sp_names,
                "symtext": sp_texts,
                "symtypes": sp_types,
                "headers": {
                    "scalar": {
                        "type": "string",
                        "alias": "Scalar Name",
                    },
                    "description": {
                        "type": "string",
                        "alias": "Scalar Description",
                    },
                    "value": {
                        "type": "string",
                        "alias": "Scalar Value",
                    },
                },
            }

        if len(ve_names) != 0:
            ve_scalars_dict = {
                "alias": "Output Variable/Equation Scalars",
                "symnames": ve_names,
                "symtext": ve_texts,
                "symtypes": ve_types,
                "headers": {
                    "scalar": {"type": "string", "alias": "Scalar Name"},
                    "description": {
                        "type": "string",
                        "alias": "Scalar Description",
                    },
                    "level": {"type": "numeric", "alias": "Level"},
                    "marginal": {"type": "numeric", "alias": "Marginal"},
                    "lower": {"type": "numeric", "alias": "Lower"},
                    "upper": {"type": "numeric", "alias": "Upper"},
                    "scale": {"type": "numeric", "alias": "Scale"},
                },
            }

        return sp_scalars_dict, ve_scalars_dict

    def validate_table(
        self, symbol: Set | Parameter | Variable | Equation, last_item
    ):
        if not gp.get_option("VALIDATION"):
            return

        if symbol.dimension < 2:
            raise ValidationError(
                f"miro table symbol `{symbol}` must have at least 2"
                " domain elements " + f"but got {symbol.dimension}."
            )

        if not isinstance(last_item, (gp.Set, gp.Alias)):
            raise ValidationError(
                f"The last domain of the miro table symbol `{symbol}` must be a set or an alias"
                f" but found {type(last_item)}"
            )

        if (
            isinstance(symbol.domain_forwarding, list)
            and symbol.domain_forwarding[-1]
        ) or (
            isinstance(symbol.domain_forwarding, bool)
            and symbol.domain_forwarding
        ):
            raise ValidationError(
                "Cannot use domain forwarding feature for miro tables."
            )

        if hasattr(last_item, "_is_miro_input") and last_item._is_miro_input:
            raise ValidationError(
                "The last column of miro table cannot be a miro input!"
            )

    def prepare_headers_dict(
        self, symbol: Set | Parameter | Variable | Equation
    ):
        if isinstance(symbol, gp.Set):
            domain_keys = symbol.domain_names + ["element_text"]
            types = ["string"] * (len(symbol.domain_names) + 1)
        elif isinstance(symbol, gp.Parameter):
            domain_keys = symbol.domain_names + ["value"]
            types = ["string"] * len(symbol.domain_names) + ["numeric"]

            if symbol._is_miro_table:
                last_item = symbol.domain[-1]
                self.validate_table(symbol, last_item)

                set_values = last_item.records["uni"].values.tolist()

                domain_keys = domain_keys[:-2]
                types = ["string"] * len(domain_keys) + ["numeric"] * len(
                    set_values
                )
                domain_keys += set_values
        elif isinstance(symbol, (gp.Variable, gp.Equation)):
            domain_keys = symbol.domain_names + [
                "level",
                "marginal",
                "lower",
                "upper",
                "scale",
            ]
            types = ["string"] * len(symbol.domain_names) + ["numeric"] * 5

        domain_values = []
        uni_counter = 0
        for idx, key in enumerate(domain_keys):
            if key == "*":
                domain_keys[idx] = (
                    "uni" if uni_counter == 0 else f"uni{uni_counter}"
                )
                uni_counter += 1

        for column, column_type in zip(domain_keys, types):
            try:
                elem = self.container[column]
                alias = elem.description if elem.description else column
            except KeyError:
                alias = column

            domain_values.append({"type": column_type, "alias": alias})

        assert len(domain_keys) == len(domain_values)
        return dict(zip(domain_keys, domain_values))

    def prepare_symbols(self, symbols: list[str]) -> list[dict]:
        type_map = {
            gp.Parameter: "parameter",
            gp.Set: "set",
            gp.Variable: "variable",
            gp.Equation: "equation",
        }

        info = []
        for name in symbols:
            symbol = self.container[name]

            headers_dict = self.prepare_headers_dict(symbol)

            info.append(
                {
                    "alias": (
                        symbol.description
                        if symbol.description
                        else symbol.name
                    ),
                    "symtype": type_map[type(symbol)],  # type: ignore
                    "headers": headers_dict,
                }
            )

        return info

    def prepare_symbols_dict(self, is_input: bool, symbols: list[str]) -> dict:
        alias = "Input Scalars" if is_input else "Output Scalars"
        scalars_sp_dict, scalars_ve_dict = self._prepare_scalars(
            alias, self.input_scalars if is_input else self.output_scalars
        )

        non_scalars = []

        for symbol in symbols:
            if is_input:
                if symbol not in self.input_scalars:
                    non_scalars.append(symbol)
            else:
                if symbol not in self.output_scalars:
                    non_scalars.append(symbol)

        symbol_dicts = self.prepare_symbols(non_scalars)

        non_scalars = [name for name in non_scalars]

        keys = non_scalars
        values = symbol_dicts

        if scalars_sp_dict is not None:
            keys.append("_scalars" if is_input else "_scalars_out")
            values.append(scalars_sp_dict)

        if scalars_ve_dict is not None:
            keys.append("_scalarsve_out")
            values.append(scalars_ve_dict)

        symbols_dict = dict(zip(keys, values))

        return symbols_dict

    def prepare_dict(self) -> dict:
        input_symbols_dict = self.prepare_symbols_dict(
            True, self.input_symbols
        )
        output_symbols_dict = self.prepare_symbols_dict(
            False, self.output_symbols
        )

        miro_dict = {
            "modelTitle": self.model_title,
            "inputSymbols": input_symbols_dict,
            "outputSymbols": output_symbols_dict,
        }

        return miro_dict

    def write_json(self):
        miro_dict = self.prepare_dict()

        filename = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        directory = os.path.dirname(sys.argv[0])
        conf_path = os.path.join(directory, f"conf_{filename}")
        try:
            os.mkdir(conf_path)
        except FileExistsError:
            pass

        with open(
            os.path.join(conf_path, f"{filename}_io.json"),
            "w",
            encoding="utf-8",
        ) as conf:
            conf.write(json.dumps(miro_dict, indent=4))

        return miro_dict
